"""
File I/O module for handling various neuroimaging file formats and associated data.

This module provides utilities for loading, validating, and processing:
- NIFTI files (.nii, .nii.gz)
- GIFTI files (.gii)
- CIFTI files (.dtseries.nii)
- Time course data (.csv, .txt)
- Task design files (.csv, .txt)
"""

from findviz.viz.io import cifti
from findviz.viz.io import gifti
from findviz.viz.io import nifti
from findviz.viz.io import timecourse
from findviz.viz.io import upload
from findviz.viz.io import validate

__all__ = [
    'cifti',
    'gifti',
    'nifti',
    'timecourse',
    'upload',
    'validate'
]


"""
Validate parameters to analysis classes
"""

from typing import List

def validate_less_than_half_time_length(
    value: int, 
    time_length: int
) -> bool:
    """
    Validate value is less than half the time length
    """
    return abs(value) <= time_length / 2


def validate_less_than_or_equal_to_zero(value: int) -> bool:
    """
    Validate value is less than or equal to zero
    """
    return value <= 0

def validate_greater_than_or_equal_to_zero(value: int) -> bool:
    """
    Validate value is greater than or equal to zero
    """
    return value >= 0




"""
Window average fmri time courses around annotation markers
"""
from typing import List

import numpy as np

from findviz.viz.analysis.utils import extract_range
from findviz.viz.analysis.validate import (
    validate_less_than_or_equal_to_zero,
    validate_greater_than_or_equal_to_zero,
    validate_less_than_half_time_length
)
from findviz.viz.exception import ParameterInputError


class WindowAverage:
    """
    Window average fmri time courses around annotation markers

    Parameters
    ----------
    left_edge : int
        left edge of window, should be zero or negative
    right_edge : int
        right edge of window, should be positive or zero
    """
    def __init__(
        self, 
        left_edge: int, 
        right_edge: int,
        n_timepoints: int
    ):
        self.left_edge = left_edge
        self.right_edge = right_edge
        self.n_timepoints = n_timepoints
        self._validate()

    def average(
        self, 
        fmri_data: np.ndarray, 
        annotation_markers: list
    ) -> np.ndarray:
        """
        Average fmri time courses around annotation markers

        Parameters
        ----------
        fmri_data : np.ndarray
            fmri data to average
        annotation_markers : list
            list of annotation markers

        Returns
        -------
        np.ndarray
            averaged fmri data
        """
        # index windows for each marker
        windows = []
        for center in annotation_markers:
            windows.append(
                extract_range(fmri_data, center, self.left_edge, self.right_edge)
            )
        # convert to 3d array
        windows = np.stack(windows, axis=0)

        # average all windows
        w_avg = np.nanmean(windows, axis=0)

        return w_avg

    def get_timepoints(self) -> List[float]:
        """
        Get timepoints for the window average

        Returns
        -------
        np.ndarray
            timepoints for the window average
        """
        return np.arange(
            self.left_edge,
            self.right_edge + 1
        ).tolist()

    def _validate(self) -> None:
        """
        Validate input parameters
        """
         # check lags are valid
        if not validate_less_than_or_equal_to_zero(self.left_edge):
            raise ParameterInputError(
                message="Left edge must be negative",
                parameters=["left_edge"]
            )
        if not validate_greater_than_or_equal_to_zero(self.right_edge):
            raise ParameterInputError(
                message="Right edge must be positive",
                parameters=["right_edge"]
            )
        if not validate_less_than_half_time_length(
            self.left_edge, 
            self.n_timepoints
        ):
            raise ParameterInputError(
                message="Left edge must be less than half the time length",
                parameters=["left_edge"]
            )
        if not validate_less_than_half_time_length(
            self.right_edge, 
            self.n_timepoints
        ):
            raise ParameterInputError(
                message="Right edge must be less than half the time length",
                parameters=["right_edge"]
            )


"""
Nifti data handling
"""
from typing import Literal, Optional, TypedDict, List, Tuple

import nibabel as nib
import numpy as np

from nilearn.image import index_img

from findviz.routes.utils import sanitize_array_for_json
from findviz.viz.viewer.types import OrthoSliceIndexDict, MontageSliceDirectionIndexDict

class SliceData(TypedDict):
    x: List[List[float]]
    y: List[List[float]]
    z: List[List[float]]

class CoordLabels(TypedDict):
    x: List[Tuple[int, int, int]]
    y: List[Tuple[int, int, int]]
    z: List[Tuple[int, int, int]]

class NiftiTimePointData(TypedDict):
    func: SliceData
    anat: Optional[SliceData]
    coord_labels: Optional[CoordLabels]


def get_nifti_data(
    time_point: int,
    func_img: nib.Nifti1Image,
    coord_labels: np.ndarray,
    slice_idx: OrthoSliceIndexDict | MontageSliceDirectionIndexDict,
    view_state: Literal['montage', 'ortho'],
    montage_slice_dir: Literal['x', 'y', 'z'],
    threshold_min: float = 0,
    threshold_max: float = 0,
    threshold_min_orig: float = 0,
    threshold_max_orig: float = 0,
    anat_img: Optional[nib.Nifti1Image] = None,
) -> NiftiTimePointData:
    """Get slice data for a specific timepoint from functional, anatomical and mask images.

    Parameters
    ----------
    time_point : int
        Index of the timepoint to extract from functional image
    func_img : nib.Nifti1Image
        4D functional image
    coord_labels : np.ndarray
        Array of voxel coordinates
    slice_idx : OrthoSliceIndexDict | MontageSliceDirectionIndexDict
        Slice indices for the current view state
    view_state : Literal['montage', 'ortho']
        Current view state of the viewer
    montage_slice_dir : Literal['x', 'y', 'z']
        Slice direction when in montage view
    threshold_min : float, optional
        Minimum threshold value, by default 0
    threshold_max : float, optional
        Maximum threshold value, by default 0
    threshold_min_orig : float, optional
        Original minimum threshold value, by default 0
    threshold_max_orig : float, optional
        Original maximum threshold value, by default 0
    anat_img : Optional[nib.Nifti1Image], optional
        3D anatomical image, by default None

    Returns
    -------
    Dict[str, Dict[str, List[List[float]]]]
        Dictionary containing slice data for each image type and axis:
        {
            'func': {'x': List[List[float]], 'y': List[List[float]], 'z': List[List[float]]},
            'anat': {'x': List[List[float]], 'y': List[List[float]], 'z': List[List[float]]},
            'mask': {'x': List[List[float]], 'y': List[List[float]], 'z': List[List[float]]},
            'coords': List[Tuple[int, int, int]]
        }
    """
    # Select time point
    func_data = np.array(index_img(func_img, time_point).get_fdata())
    
    # threshold data if threshold_min or threshold_max have been changed
    if (threshold_min != threshold_min_orig) or (threshold_max != threshold_max_orig):
        func_data = threshold_nifti_data(
            nifti_data=func_data, 
            threshold_min=threshold_min, 
            threshold_max=threshold_max
        )

    # get anatomical data if present
    anat_data = anat_img.get_fdata() if anat_img else None
    
    # initialize output
    slice_out = {
        'func': {},
        'anat': {},
        'coords': {}
    }
    # Loop through axes and update
    for axis, slice_container in zip(
        ['x', 'y', 'z'], 
        ['slice_1', 'slice_2', 'slice_3']
    ):
        # set axis to all same direction, if montage
        if view_state == 'montage':
            nifti_axis = montage_slice_dir
            # get slice idx
            slice_i = slice_idx[slice_container][montage_slice_dir]
        else:
            nifti_axis = axis
            # get slice idx
            slice_i = slice_idx[axis]

        slice_out['func'][slice_container] = get_slice_data(
            func_data, slice_i, axis=nifti_axis
        )
        if anat_img is not None:
            slice_out['anat'][slice_container] = get_slice_data(
                anat_data, slice_i, axis=nifti_axis
            )
        # get coord labels
        slice_out['coords'][slice_container] = get_slice_data(
            coord_labels, slice_i, axis=nifti_axis, array_type='string'
        )

    return slice_out


def get_slice_data(
    nifti_data: np.ndarray,
    slice_index: int, 
    axis: Literal['x', 'y', 'z'],
    array_type: Literal['number', 'string'] = 'number'
) -> List[List[float]]:
    """Extract a 2D slice from a NIfTI image along a specified axis.

    Parameters
    ----------
    nifti_data : np.ndarray
        Input NIfTI data
    slice_index : int
        Index of the slice to extract
    axis : Literal['x', 'y', 'z']
        Axis along which to take the slice
    array_type : Literal['number', 'string']
        The type of elements in the input array
    Returns
    -------
    2D numpy array containing the slice data, transposed for display
    """
    # Extract the slice data
    if axis == 'x':
        slice_data = np.squeeze(
            nifti_data[slice_index, :, :]
        ).transpose()
    elif axis == 'y':
        slice_data = np.squeeze(
            nifti_data[:, slice_index, :]
        ).transpose()
    elif axis == 'z':
        slice_data = np.squeeze(
            nifti_data[:, :, slice_index]
        ).transpose()

    # convert to list if string
    if array_type == 'string':
        return slice_data.tolist()
    # otherwise sanitize for json (e.g. handle NaN values)
    else:
        return sanitize_array_for_json(slice_data)


def get_timecourse_nifti(
    func_img: nib.Nifti1Image,
    x: int,
    y: int,
    z: int,
) -> Tuple[List[float], str]:
    """
    Get timecourse data and voxel label for a specific voxel 
    from a functional NIFTI file.

    Parameters
    ----------
    func_img : nib.Nifti1Image
        Functional NIFTI image
    x : int
        X-coordinate of the voxel
    y : int
        Y-coordinate of the voxel
    z : int
        Z-coordinate of the voxel
    """
    # create time course label
    time_course_label = f'Voxel: (x={x}, y={y}, z={z})'
    time_course = func_img.get_fdata()[x, y, z, :].tolist()
    return time_course, time_course_label


def threshold_nifti_data(
    nifti_data: np.ndarray,
    threshold_min: float,
    threshold_max: float,
) -> np.ndarray: 
    """Threshold a NIfTI 3d array

    Parameters
    ----------
    nifti_data : np.ndarray
        Nifti 3d array to threshold
    threshold_min : float
        Minimum threshold value
    threshold_max : float
        Maximum threshold value

    Returns
    -------
    np.ndarray
        Thresholded array. Voxels with values outside 
        the threshold range are set to None.
    """
    # create mask
    mask = (nifti_data >= threshold_min) & (nifti_data <= threshold_max)
    # apply mask
    nifti_data[mask] = np.nan
    return nifti_data




"""
Plot routes
Routes:
    ADD_ANNOTATION_MARKER: Add annotation marker
    CHANGE_TASK_CONVOLUTION: Change task convolution
    CHECK_FMRI_PREPROCESSED: Check if fmri data is preprocessed
    CHECK_TS_PREPROCESSED: Check if timecourse is preprocessed
    CLEAR_ANNOTATION_MARKERS: Clear annotation markers
    GET_ANNOTATION_MARKERS: Get annotation markers
    GET_ANNOTATION_MARKER_PLOT_OPTIONS: Get annotation marker plot options
    GET_DISTANCE_PLOT_OPTIONS: Get distance plot options
    GET_FMRI_PLOT_OPTIONS: Get fMRI plot options
    GET_NIFTI_VIEW_STATE: Get nifti view state (ortho or montage)
    GET_TASK_DESIGN_PLOT_OPTIONS: Get task design plot options
    GET_TIMECOURSE_GLOBAL_PLOT_OPTIONS: Get timecourse global plot options
    GET_TIMECOURSE_PLOT_OPTIONS: Get timecourse plot options
    GET_TIMEMARKER_PLOT_OPTIONS: Get timemarker plot options
    GET_TS_FMRI_PLOTTED: Get whether an fmri timecourse is plotted
    MOVE_ANNOTATION_SELECTION: Move annotation selection
    RESET_FMRI_COLOR_OPTIONS: Reset fMRI color options
    REMOVE_DISTANCE_PLOT: Remove distance plot
    UPDATE_ANNOTATION_MARKER_PLOT_OPTIONS: Update annotation marker plot options
    UPDATE_DISTANCE_PLOT_OPTIONS: Update distance plot options
    UPDATE_FMRI_PLOT_OPTIONS: Update fMRI plot options
    UPDATE_TIMECOURSE_GLOBAL_PLOT_OPTIONS: Update timecourse global plot options
    UPDATE_TIMECOURSE_PLOT_OPTIONS: Update timecourse plot options
    UPDATE_TIMECOURSE_SHIFT: Update timecourse shift (scale or constant)
    UPDATE_TIMEMARKER_PLOT_OPTIONS: Update timemarker plot options
    UPDATE_NIFTI_VIEW_STATE: Update nifti view state (ortho or montage)
"""

import json

from flask import Blueprint, request

from findviz.logger_config import setup_logger
from findviz.routes.utils import (
    convert_value,
    handle_context,
    handle_route_errors,
    Routes
)
from findviz.routes.shared import data_manager

logger = setup_logger(__name__)
plot_bp = Blueprint('plot', __name__)

@plot_bp.route(Routes.ADD_ANNOTATION_MARKER.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in add annotation marker request',
    log_msg='Added annotation marker successfully',
    route_parameters=['marker'],
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.ADD_ANNOTATION_MARKER
)
def add_annotation_marker() -> dict:
    """Add annotation marker"""
    marker = convert_value(request.form['marker'])
    data_manager.ctx.add_annotation_markers(marker)
    return {'marker': marker}


@plot_bp.route(Routes.CHANGE_TASK_CONVOLUTION.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in change task convolution request',
    log_msg='Changed task convolution successfully',
    route_parameters=['convolution'],
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.CHANGE_TASK_CONVOLUTION
)
def change_task_convolution() -> int:
    """Change task convolution globally across all tasks"""
    convolution = convert_value(request.form['convolution'])
    data_manager.ctx.update_timecourse_global_plot_options(
        {'global_convolution': convolution}
    )
    if convolution:
        conv_type = 'hrf'
    else:
        conv_type = 'block'
    # update all task design plot options
    for label in data_manager.ctx.task_plot_options:
        data_manager.ctx.update_task_design_plot_options(
            label, {'convolution': conv_type}
        )
    return {'status': 'success'}


@plot_bp.route(Routes.CHECK_FMRI_PREPROCESSED.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in check fmri preprocessed request',
    log_msg='Checked fmri preprocessed successfully',
    route=Routes.CHECK_FMRI_PREPROCESSED
)
def check_fmri_preprocessed() -> dict:
    """Check if fmri data is preprocessed"""
    return {'is_preprocessed': data_manager.ctx.fmri_preprocessed}


@plot_bp.route(Routes.CHECK_TS_PREPROCESSED.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in check timecourse preprocessed request',
    log_msg='Checked timecourse preprocessed successfully',
    route_parameters=['label', 'ts_type'],
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.CHECK_TS_PREPROCESSED
)
def check_ts_preprocessed() -> dict:
    """Check if timecourse is preprocessed"""
    label = convert_value(request.form['label'])
    ts_type = convert_value(request.form['ts_type'])
    if ts_type == 'timecourse':
        is_preprocessed = data_manager.ctx.check_ts_preprocessed(label)
    else:
        # task design is not preprocessed
        is_preprocessed = False

    return {'is_preprocessed': is_preprocessed}


@plot_bp.route(Routes.CLEAR_ANNOTATION_MARKERS.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in clear annotation markers request',
    log_msg='Cleared annotation markers successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.CLEAR_ANNOTATION_MARKERS
)
def clear_annotation_markers() -> dict:
    """Clear annotation markers"""
    data_manager.ctx.clear_annotation_markers()
    return {'status': 'success'}


@plot_bp.route(Routes.GET_ANNOTATION_MARKERS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in annotation markers request',
    log_msg='Retrieved annotation markers successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_ANNOTATION_MARKERS
)
def get_annotation_markers() -> dict:
    """Get annotation markers, annotation selection, and annotation plot options"""
    markers = data_manager.ctx.annotation_markers
    selection = data_manager.ctx.annotation_selection
    plot_options = data_manager.ctx.annotation_marker_plot_options.to_dict()
    return {
        'markers': markers,
        'selection': selection,
        'plot_options': plot_options
    }


@plot_bp.route(Routes.GET_ANNOTATION_MARKER_PLOT_OPTIONS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in get annotation marker plot options request',
    log_msg='Retrieved annotation marker plot options successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_ANNOTATION_MARKER_PLOT_OPTIONS
)
def get_annotation_marker_plot_options() -> dict:
    """Get annotation marker plot options"""
    return data_manager.ctx.get_annotation_marker_plot_options()


@plot_bp.route(Routes.GET_DISTANCE_PLOT_OPTIONS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in distance plot options request',
    log_msg='Retrieved distance plot options successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_DISTANCE_PLOT_OPTIONS
)
def get_distance_plot_options() -> dict:
    """Get current distance plot options."""
    return data_manager.ctx.get_distance_plot_options()


@plot_bp.route(Routes.GET_FMRI_PLOT_OPTIONS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in get fMRI plot options request',
    log_msg='Retrieved fMRI plot options successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_FMRI_PLOT_OPTIONS
)
def get_fmri_plot_options() -> dict:
    """Get current fMRI plot options."""
    return data_manager.ctx.get_fmri_plot_options()


@plot_bp.route(Routes.GET_NIFTI_VIEW_STATE.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in nifti view state request',
    log_msg='Retrieved nifti view state successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_NIFTI_VIEW_STATE
)
def get_nifti_view_state() -> dict:
    """Get current nifti view state"""
    return {'view_state': data_manager.ctx.view_state}


@plot_bp.route(Routes.GET_TASK_DESIGN_PLOT_OPTIONS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in task design plot options request',
    log_msg='Retrieved task design plot options successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_TASK_DESIGN_PLOT_OPTIONS,
    route_parameters=['label']
)
def get_task_design_plot_options() -> dict:
    """Get current task design plot options"""
    label = convert_value(request.args.get('label'))
    return data_manager.ctx.get_task_design_plot_options(label)


@plot_bp.route(Routes.GET_TIMECOURSE_GLOBAL_PLOT_OPTIONS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in timecourse global plot options request',
    log_msg='Retrieved timecourse global plot options successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_TIMECOURSE_GLOBAL_PLOT_OPTIONS
)
def get_timecourse_global_plot_options() -> dict:
    """Get current timecourse global plot options"""
    return data_manager.ctx.get_timecourse_global_plot_options()


@plot_bp.route(Routes.GET_TIMECOURSE_PLOT_OPTIONS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in timecourse plot options request',
    log_msg='Retrieved timecourse plot options successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_TIMECOURSE_PLOT_OPTIONS,
    route_parameters=['label']
)
def get_timecourse_plot_options() -> dict:
    """Get current timecourse plot options"""
    label = convert_value(request.args.get('label'))
    return data_manager.ctx.get_timecourse_plot_options(label)


@plot_bp.route(Routes.GET_TIMECOURSE_SHIFT_HISTORY.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in timecourse shift history request',
    log_msg='Retrieved timecourse shift history successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_TIMECOURSE_SHIFT_HISTORY,
    route_parameters=['label', 'source']
)
def get_timecourse_shift_history() -> dict:
    """Get current timecourse shift history"""
    label = convert_value(request.args.get('label'))
    source = convert_value(request.args.get('source'))
    return data_manager.ctx.get_timecourse_shift_history(label, source)


@plot_bp.route(Routes.GET_TIMEMARKER_PLOT_OPTIONS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in timemarker plot options request',
    log_msg='Retrieved timemarker plot options successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_TIMEMARKER_PLOT_OPTIONS
)
def get_timemarker_plot_options() -> dict:
    """Get current timemarker plot options"""
    return data_manager.ctx.get_time_marker_plot_options()


@plot_bp.route(Routes.GET_TS_FMRI_PLOTTED.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in get ts fmri plotted request',
    log_msg='Retrieved ts fmri plotted successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_TS_FMRI_PLOTTED
)
def get_ts_fmri_plotted() -> dict:
    """Get whether an fmri timecourse is plotted"""
    return {'ts_fmri_plotted': data_manager.ctx.ts_fmri_plotted}


@plot_bp.route(Routes.MOVE_ANNOTATION_SELECTION.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in move annotation selection request',
    log_msg='Moved annotation selection successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.MOVE_ANNOTATION_SELECTION
)
def move_annotation_selection() -> dict:
    """Move annotation selection"""
    direction = convert_value(request.form['direction'])
    selected_marker = data_manager.ctx.move_annotation_selection(direction)
    return {'selected_marker': selected_marker}


@plot_bp.route(Routes.REMOVE_DISTANCE_PLOT.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in remove distance plot request',
    log_msg='Removed distance plot successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.REMOVE_DISTANCE_PLOT
)
def remove_distance_plot() -> dict:
    """Remove distance plot"""
    data_manager.ctx.clear_distance_plot_state()
    return {'status': 'success'}


@plot_bp.route(Routes.RESET_FMRI_COLOR_OPTIONS.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in reset fMRI color options request',
    log_msg='Reset fMRI color options successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.RESET_FMRI_COLOR_OPTIONS
)
def reset_fmri_color_options() -> dict:
    """Reset fMRI color options to defaults."""
    data_manager.ctx.reset_fmri_color_options()
    return {'status': 'success'}


@plot_bp.route(Routes.RESET_TIMECOURSE_SHIFT.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in reset timecourse shift request',
    log_msg='Reset timecourse shift successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.RESET_TIMECOURSE_SHIFT,
    route_parameters=['label', 'source', 'change_type']
)
def reset_timecourse_shift() -> dict:
    """Reset timecourse shift"""
    label = convert_value(request.form['label'])
    change_type = convert_value(request.form['change_type'])
    source = convert_value(request.form['source'])
    data_manager.ctx.reset_timecourse_shift(label, change_type, source)
    return {'status': 'success'}


@plot_bp.route(Routes.UNDO_ANNOTATION_MARKER.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in undo annotation marker request',
    log_msg='Undid annotation marker successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.UNDO_ANNOTATION_MARKER
)
def undo_annotation_marker() -> dict:
    """Undo annotation marker"""
    data_manager.ctx.pop_annotation_marker()
    # get current annotation markers
    markers = data_manager.ctx.annotation_markers
    # check if there are any markers
    if len(markers) > 0:
        # get most recent marker
        marker = markers[-1]
    else:
        marker = None
    return {'marker': marker}


@plot_bp.route(Routes.UPDATE_DISTANCE_PLOT_OPTIONS.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in distance plot options update request',
    log_msg='Updated distance plot options successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.UPDATE_DISTANCE_PLOT_OPTIONS,
    route_parameters=['distance_plot_options']
)
def update_distance_plot_options() -> dict:
    """Update distance plot options"""
    distance_plot_options = json.loads(request.form['distance_plot_options'])
    data_manager.ctx.update_distance_plot_options(distance_plot_options)
    return {'status': 'success'}


@plot_bp.route(Routes.UPDATE_FMRI_PLOT_OPTIONS.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in fMRI plot options update request',
    log_msg='Updated fMRI plot options successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.UPDATE_FMRI_PLOT_OPTIONS,
    route_parameters=['fmri_plot_options']
)
def update_fmri_plot_options() -> dict:
    """Update plot options based on form data."""
    fmri_plot_options = json.loads(request.form['fmri_plot_options'])
    data_manager.ctx.update_fmri_plot_options(fmri_plot_options)
    return {'status': 'success'}


@plot_bp.route(Routes.UPDATE_ANNOTATION_MARKER_PLOT_OPTIONS.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in update annotation marker plot options request',
    log_msg='Updated annotation marker plot options successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.UPDATE_ANNOTATION_MARKER_PLOT_OPTIONS,
    route_parameters=['annotation_marker_plot_options']
)
def update_annotation_marker_plot_options() -> dict:
    """Update annotation marker plot options"""
    annotation_marker_plot_options = json.loads(
        request.form['annotation_marker_plot_options']
    )
    data_manager.ctx.update_annotation_marker_plot_options(
        annotation_marker_plot_options
    )
    return {'status': 'success'}


@plot_bp.route(Routes.UPDATE_NIFTI_VIEW_STATE.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in nifti view state update request',
    log_msg='Updated nifti view state successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.UPDATE_NIFTI_VIEW_STATE,
    route_parameters=['view_state']
)
def update_nifti_view_state() -> dict:
    """Update nifti view state"""
    view_state = convert_value(request.form['view_state'])
    data_manager.ctx.update_view_state(view_state)
    return {'status': 'success'}


@plot_bp.route(Routes.UPDATE_TASK_DESIGN_PLOT_OPTIONS.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in task design plot options update request',
    log_msg='Updated task design plot options successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.UPDATE_TASK_DESIGN_PLOT_OPTIONS,
    route_parameters=['label']
)
def update_task_design_plot_options() -> dict:
    """Update task design plot options"""
    label = convert_value(request.form['label'])
    task_design_plot_options = json.loads(request.form['task_design_plot_options'])
    data_manager.ctx.update_task_design_plot_options(label, task_design_plot_options)
    return {'status': 'success'}


@plot_bp.route(Routes.UPDATE_TIMECOURSE_GLOBAL_PLOT_OPTIONS.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in timecourse global plot options update request',
    log_msg='Updated timecourse global plot options successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.UPDATE_TIMECOURSE_GLOBAL_PLOT_OPTIONS,
    route_parameters=['timecourse_global_plot_options']
)
def update_timecourse_global_plot_options() -> dict:
    """Update timecourse global plot options"""
    timecourse_global_plot_options = json.loads(
        request.form['timecourse_global_plot_options']
    )
    data_manager.ctx.update_timecourse_global_plot_options(
        timecourse_global_plot_options
    )
    return {'status': 'success'}


@plot_bp.route(Routes.UPDATE_TIMECOURSE_PLOT_OPTIONS.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in timecourse plot options update request',
    log_msg='Updated timecourse plot options successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.UPDATE_TIMECOURSE_PLOT_OPTIONS,
    route_parameters=['label', 'timecourse_plot_options']
)
def update_timecourse_plot_options() -> dict:
    """Update timecourse plot options"""
    label = convert_value(request.form['label'])
    timecourse_plot_options = json.loads(request.form['timecourse_plot_options'])
    # convert values
    timecourse_plot_options = {
        convert_value(key): convert_value(value)
        for key, value in timecourse_plot_options.items()
    }
    data_manager.ctx.update_timecourse_plot_options(label, timecourse_plot_options)
    return {'status': 'success'}


@plot_bp.route(Routes.UPDATE_TIMECOURSE_SHIFT.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in timecourse shift update request',
    log_msg='Updated timecourse shift successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.UPDATE_TIMECOURSE_SHIFT,
    route_parameters=['label', 'source', 'change_type', 'change_direction']
)
def update_timecourse_shift() -> dict:
    """Update timecourse shift"""
    label = convert_value(request.form['label'])
    source = convert_value(request.form['source'])
    change_type = convert_value(request.form['change_type'])
    change_direction = convert_value(request.form['change_direction'])
    data_manager.ctx.update_timecourse_shift(label, source, change_type, change_direction)
    return {'status': 'success'}


@plot_bp.route(Routes.UPDATE_TIMEMARKER_PLOT_OPTIONS.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in timemarker plot options update request',
    log_msg='Updated timemarker plot options successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.UPDATE_TIMEMARKER_PLOT_OPTIONS,
    route_parameters=['timemarker_plot_options']
)
def update_timemarker_plot_options() -> dict:
    """Update timemarker plot options"""
    timemarker_plot_options = json.loads(request.form['timemarker_plot_options'])
    data_manager.ctx.update_time_marker_plot_options(timemarker_plot_options)
    return {'status': 'success'}




"""Flask routes"""
from findviz.routes import file
from findviz.routes.viewer import color, io, plot, preprocess, data
from findviz.routes import utils


__all__ = [
    'color',
    'data',
    'file',
    'io',
    'plot',
    'preprocess',
    'utils',
]



"""
Types for viewer
"""

from typing import TypedDict, Optional, List, Dict, Literal, Tuple

import nibabel as nib
import numpy as np

class AnnotationMarkerPlotOptionsDict(TypedDict):
    """Dictionary of marker plot options"""
    opacity: Optional[float]
    width: Optional[float]
    shape: Optional[Literal['solid', 'dashed', 'dotted']]
    color: Optional[str]
    highlight: Optional[bool]

class CrosshairCoordsDict(TypedDict):
    """Dictionary of crosshair coordinates for nifti data plotting"""
    slice1: Dict[Literal['len_x', 'len_y', 'x', 'y'], int]
    slice2: Dict[Literal['len_x', 'len_y', 'x', 'y'], int]
    slice3: Dict[Literal['len_x', 'len_y', 'x', 'y'], int]

class DirectionLabelCoordsDict(TypedDict):
    """Dictionary of direction label coordinates for nifti data plotting"""
    slice1: Dict[str, Dict[str, int]]
    slice2: Dict[str, Dict[str, int]]
    slice3: Dict[str, Dict[str, int]]

class DistancePlotOptionsDict(TypedDict):
    """Output dict from get_distance_plot_options() method"""
    color_map: str
    color_min: float
    color_max: float
    color_range: Tuple[float, float]
    precision: int
    allowed_precision: int
    slider_step_size: float
    time_marker_on: bool
    time_marker_color: str
    time_marker_width: float
    time_marker_opacity: float

class FmriPlotOptionsDict(TypedDict):
    """Output dict from get_fmri_plot_options() method"""
    color_min: Optional[float]
    color_max: Optional[float]
    color_range: Optional[Tuple[float, float]]
    opacity: float
    threshold_min: float
    threshold_max: float
    threshold_range: Optional[Tuple[float, float]]
    color_map: str
    hover_text_on: bool
    precision: int
    slider_steps: int
    allowed_precision: int
    crosshair_on: bool
    direction_marker_on: bool

class GiftiDataDict(TypedDict):
    """Dictionary of GIFTI data input"""
    left_func_img: nib.GiftiImage
    right_func_img: nib.GiftiImage

class GiftiDataPreprocessedDict(TypedDict):
    """Dictionary of preprocessed GIFTI data input"""
    left_func_img: nib.GiftiImage
    right_func_img: nib.GiftiImage

class ColorOptions(TypedDict):
    """Color options for fmri plot"""
    color_min: float
    color_max: float
    threshold_min: float
    threshold_max: float
    opacity: float

class SliceCoordsDict(TypedDict):
    """Dictionary of x,y click coordinates (integers) for nifti slices"""
    slice_1: Dict[Literal['x', 'y'], int]
    slice_2: Dict[Literal['x', 'y'], int]
    slice_3: Dict[Literal['x', 'y'], int]

class OrthoSliceIndexDict(TypedDict):
    """Dictionary of x,y,z slice indices for ortho view"""
    x: int
    y: int
    z: int

class MontageSliceDirectionIndexDict(TypedDict):
    """Dictionary of x,y,z slice indices for a direction in montage view"""
    slice_1: Dict[Literal['x', 'y', 'z'], int]
    slice_2: Dict[Literal['x', 'y', 'z'], int]
    slice_3: Dict[Literal['x', 'y', 'z'], int]

class MontageSliceIndexDict(TypedDict):
    """Dictionary of montage slice indices for montage view
    Keys are slice directions:
    x: axial
    y: coronal
    z: sagittal

    Each slice direction has a dictionary of indices for the three slices.
    """
    x: MontageSliceDirectionIndexDict # axial
    y: MontageSliceDirectionIndexDict # coronal
    z: MontageSliceDirectionIndexDict # sagittal

class MontageSliceCoordsDict(TypedDict):
    """Dictionary of montage slice click coordinates (integers)
    Keys are slice directions:
    x: axial
    y: coronal
    z: sagittal

    Each slice direction has a dictionary of coordinates for the three slices.
    """
    x: SliceCoordsDict # axial
    y: SliceCoordsDict # coronal
    z: SliceCoordsDict # sagittal

class NiftiDataDict(TypedDict):
    """Dictionary of NIFTI data input"""
    func_img: nib.Nifti1Image
    anat_img: nib.Nifti1Image
    mask_img: nib.Nifti1Image

class NiftiDataPreprocessedDict(TypedDict):
    """Dictionary of preprocessed NIFTI data input"""
    func_img: nib.Nifti1Image

class TaskDesignPlotOptionsDict(TypedDict):
    """Output dict from get_task_design_plot_options() method"""
    convolution: Optional[bool]
    scale: Optional[float]
    color: Optional[str]
    width: Optional[float]
    opacity: Optional[float]
    mode: Optional[Literal['lines', 'markers', 'lines+markers']]

# define output dicts from get_timecourse_global_plot_options() method
class TimeCourseGlobalPlotOptionsDict(TypedDict):
    """Output dict from get_timecourse_global_plot_options() method"""
    grid_on: Optional[bool]
    hover_text_on: Optional[bool]
    time_marker_on: Optional[bool]
    global_convolution: Optional[bool]
    
# define output dicts from get_timecourse_plot_options() method
class TimeCoursePlotOptionsDict(TypedDict):
    """Output dict from get_timecourse_plot_options() method"""
    color: Optional[str]
    plotType: Optional[Literal['block', 'hrf']]
    width: Optional[float]
    opacity: Optional[float]
    mode: Optional[Literal['lines', 'markers', 'lines+markers']]
    scale: Optional[float]

# define output dicts from get_time_marker_plot_options() method
class TimeMarkerPlotOptionsDict(TypedDict):
    """output dict from get_time_marker_plot_options() method"""
    opacity: Optional[float]
    width: Optional[float]
    shape: Optional[Literal['solid', 'dashed', 'dotted']]
    color: Optional[str]

# define outpus from get_metadata() method
class ViewerMetadataNiftiDict(TypedDict):
    """output dict from get_metadata() method for nifti data"""
    file_type: Literal['nifti']
    timepoint: int
    anat_input: bool
    mask_input: bool
    tr: Optional[float]
    slicetime_ref: Optional[float]
    view_state: Literal['ortho', 'montage']
    montage_slice_dir: Literal['x', 'y', 'z']
    timepoints: List[int]
    fmri_preprocessed: bool
    ts_preprocessed: bool
    global_min: float
    global_max: float
    slice_len: Dict[str, int]
    ts_enabled: bool
    task_enabled: bool  

class ViewerMetadataGiftiDict(TypedDict):
    """output dict from get_metadata() method for gifti data"""
    file_type: Literal['gifti']
    timepoint: int
    left_input: bool
    right_input: bool
    tr: Optional[float]
    slicetime_ref: Optional[float]
    vertices_left: List[float]
    faces_left: List[float]
    vertices_right: List[float]
    faces_right: List[float]
    timepoints: List[int]
    fmri_preprocessed: bool
    ts_preprocessed: bool
    global_min: float
    global_max: float
    ts_enabled: bool
    task_enabled: bool


# define output dicts from get_viewer_data() method
class ViewerDataNiftiDict(TypedDict):
    """output dict from get_viewer_data() method for nifti data"""
    anat_input: Optional[bool]
    mask_input: Optional[bool]
    func_data: Optional[np.ndarray]
    anat_data: Optional[np.ndarray]
    mask_data: Optional[np.ndarray]
    ts: Optional[Dict[str, List[float]]]
    task: Optional[Dict[str, List[float]]]
    is_fmri_preprocessed: Optional[bool]
    is_ts_preprocessed: Optional[bool]
    coord_labels: Optional[List[Tuple[int, int, int]]]

class ViewerDataGiftiDict(TypedDict):
    """output dict from get_viewer_data() method for gifti data"""
    left_input: Optional[bool]
    right_input: Optional[bool]
    func_data: Optional[np.ndarray]
    ts: Optional[Dict[str, List[float]]]
    task: Optional[Dict[str, List[float]]]
    is_fmri_preprocessed: Optional[bool]
    is_ts_preprocessed: Optional[bool]



"""
This module defines the plot options classes for the FIND viewer.
"""

from enum import Enum
from dataclasses import dataclass, field
from typing import Literal, Optional, Dict, Tuple

from findviz.viz.analysis.scaler import SignalScaler, SignalShifter

class ColorMaps(Enum):
    """
    Enum for color maps
    """
    GREYS = 'Greys'
    YLGNBU = 'YlGnBu'
    GREENS = 'Greens'
    YLORRD = 'YlOrRd'
    BLUERED = 'Bluered'
    RDBU = 'RdBu'
    REDS = 'Reds'
    BLUES = 'Blues'
    PICNIC = 'Picnic'
    RAINBOW = 'Rainbow'
    PORTLAND = 'Portland'
    JET = 'Jet'
    HOT = 'Hot'
    BLACKBODY = 'Blackbody'
    ELECTRIC = 'Electric'
    VIRIDIS = 'Viridis'
    CIVIDIS = 'Cividis'

class TimeCourseColor(Enum):
    """Time course color options."""
    RED = 'red'
    BLUE = 'blue'
    GREEN = 'green'
    ORANGE = 'orange'
    PURPLE = 'purple'
    BLACK = 'black'
    GREY = 'grey'
    YELLOW = 'yellow'
    BROWN = 'brown'
    DARKRED = 'darkred'
    TOMATO = 'tomato'
    MAGENTA = 'magenta'
    CRIMSON = 'crimson'
    SALMON = 'salmon'
    FIREBRICK = 'firebrick'
    NAVY = 'navy'
    ROYALBLUE = 'royalblue'
    TURQUOISE = 'turquoise'
    CORAL = 'coral'
    SEA_GREEN = 'seagreen'
    FOREST_GREEN = 'forestgreen'
    OLIVE = 'olive'
    DARK_GREEN = 'darkgreen'
    DARK_ORANGE = 'darkorange'
    VIOLET = 'violet'
    SLATE_GREY = 'slategrey'
    SLATE_BLUE = 'slateblue'
    DARK_GREY = 'darkgrey'
    KHAKI = 'khaki'
    TAN = 'tan'


@dataclass
class AnnotationMarkerPlotOptions:
    """Marker plot options.
    
    Attributes:
        opacity: Opacity of the time markers. Default is 0.8
        width: Width of the time markers. Default is 1.5
        shape: Shape of the time markers. Default is 'solid'
        color: Color of the time markers. Default is RED
        highlight: Whether to highlight the time markers. Default is True
    """
    opacity: float = 0.8
    width: float = 1.5
    shape: Literal['solid', 'dash', 'dot', 'longdash', 'dashdot'] = 'solid'
    color: TimeCourseColor = TimeCourseColor.RED
    highlight: bool = True

    def to_dict(self) -> Dict[str, float]:
        """Convert to dictionary."""
        return {
            'opacity': self.opacity,
            'width': self.width,
            'shape': self.shape,
            'color': self.color.value,
            'highlight': self.highlight
        }
    
    def update_from_dict(self, data: Dict[str, float]) -> None:
        """Update from dictionary."""
        for key, value in data.items():
            # handle enum values
            if key == 'color':
                setattr(self, key, TimeCourseColor(value))
            else:
                setattr(self, key, value)
    
@dataclass
class DistancePlotOptions:
    """Distance plot options.

    Attributes:
        color_min: Minimum value for color mapping. Default is None
        color_max: Maximum value for color mapping. Default is None
        color_range: Color range for color mapping. Default is None
        color_map: Color map for color mapping. Default is 'RdBu'
        precision: Precision of the color mapping. Default is 6
        slider_step_size: Stepsize of the sliders. Default is 1.0
        allowed_precision: Allowed precision of the sliders. Default is 6
        time_marker_on: Whether time marker is enabled. Default is True
        time_marker_color: Color of the time marker. Default is BLACK
        time_marker_width: Width of the time marker. Default is 1.0
        time_marker_opacity: Opacity of the time marker. Default is 0.8
    """
    color_min: Optional[float] = None
    color_max: Optional[float] = None
    color_range: Optional[Tuple[float, float]] = None
    color_map: ColorMaps = ColorMaps.RDBU
    precision: int = 6
    slider_step_size: float = 1.0
    allowed_precision: int = 6
    time_marker_on: bool = True
    time_marker_color: TimeCourseColor = TimeCourseColor.BLACK
    time_marker_width: float = 1.0
    time_marker_opacity: float = 0.8

    def to_dict(self) -> Dict[str, float]:
        """Convert to dictionary."""
        return {
            'color_map': self.color_map.value,
            'color_min': self.color_min,
            'color_max': self.color_max,
            'color_range': self.color_range,
            'precision': self.precision,
            'allowed_precision': self.allowed_precision,
            'slider_step_size': self.slider_step_size,
            'time_marker_on': self.time_marker_on,
            'time_marker_color': self.time_marker_color.value,
            'time_marker_width': self.time_marker_width,
            'time_marker_opacity': self.time_marker_opacity
        }
    
    def update_from_dict(self, data: Dict[str, any]) -> None:
        """Update from dictionary."""
        for key, value in data.items():
            # handle enum values
            if key == 'color_map':
                setattr(self, key, ColorMaps(value))
            elif key == 'time_marker_color':
                setattr(self, key, TimeCourseColor(value))
            else:
                setattr(self, key, value)

@dataclass
class FmriPlotOptions:
    """Plot options for fMRI data.
    
    Attributes:
        color_min: Minimum value for color mapping. Default is None
        color_max: Maximum value for color mapping. Default is None
        color_range: Color range for color mapping. Default is None
        colorbar_on: Whether colorbar is enabled. Default is True
        opacity: Opacity of the color mapping. Default is 1
        threshold_min: Minimum value for threshold mapping. Default is 0
        threshold_max: Maximum value for threshold mapping. Default is 0
        threshold_range: Threshold range for threshold mapping. Default is None
        tr_convert_on: Whether TR conversion is enabled. Default is False
        color_map: Color map for color mapping. Default is 'Viridis'
        reverse_colormap: Whether the colormap is reversed. Default is False
        hover_text_on: Whether hover text is enabled. Default is True
        precision: Precision of the color mapping. Default is 6
        play_movie_speed: Speed of the movie (in milliseconds). Default is 500
        slider_step_size: Stepsize of the sliders. Default is 100
        allowed_precision: Allowed precision of the sliders. Default is 6
        view_state: View state ('ortho' or 'montage') for NIFTI data. 
            Default is 'ortho'
        crosshair_on: Whether crosshair is enabled for NIFTI data. 
            Default is True
        direction_marker_on: Whether direction marker is enabled for NIFTI data. 
            Default is False
        freeze_view_on: Whether freeze view is enabled for GIFTI data. 
            Default is False
        fmri_timecourse_enabled: Whether fmri timecourse plottingis enabled. 
            Default is False
        fmri_timecourse_freeze: Whether fmri timecourse plot selections are frozen. 
            Default is False
    """
    # fmri plot options
    color_min: Optional[float] = None
    color_max: Optional[float] = None
    color_range: Optional[Tuple[float, float]] = None
    colorbar_on: bool = True
    opacity: float = 1
    threshold_min: float = 0
    threshold_max: float = 0
    threshold_range: Optional[Tuple[float, float]] = None
    tr_convert_on: bool = False
    color_map: ColorMaps = ColorMaps.VIRIDIS
    reverse_colormap: bool = False
    hover_text_on: bool = True
    precision: int = 6
    play_movie_speed: int = 500
    slider_step_size: int = 100
    allowed_precision: int = 6
    crosshair_on: bool = True
    direction_marker_on: bool = False
    freeze_view_on: bool = False
    fmri_timecourse_enabled: bool = False
    fmri_timecourse_freeze: bool = False

    def to_dict(self) -> Dict[str, float]:
        """Convert to dictionary."""
        return {
            'color_map': self.color_map.value,
            'reverse_colormap': self.reverse_colormap,
            'color_min': self.color_min,
            'color_max': self.color_max,
            'color_range': self.color_range,
            'colorbar_on': self.colorbar_on,
            'threshold_min': self.threshold_min,
            'threshold_max': self.threshold_max,
            'threshold_range': self.threshold_range,
            'tr_convert_on': self.tr_convert_on,
            'opacity': self.opacity,
            'hover_text_on': self.hover_text_on,
            'precision': self.precision,
            'play_movie_speed': self.play_movie_speed,
            'slider_step_size': self.slider_step_size,
            'allowed_precision': self.allowed_precision,
            'crosshair_on': self.crosshair_on,
            'direction_marker_on': self.direction_marker_on,
            'freeze_view_on': self.freeze_view_on,
            'fmri_timecourse_enabled': self.fmri_timecourse_enabled,
            'fmri_timecourse_freeze': self.fmri_timecourse_freeze
        }
    

    def update_from_dict(self, data: Dict[str, any]) -> None:
        """Update from dictionary."""
        for key, value in data.items():
            # handle enum values
            if key == 'color_map':
                setattr(self, key, ColorMaps(value))
            else:
                setattr(self, key, value)

@dataclass
class TimeCourseGlobalPlotOptions:
    """Global time course plot options.
    
    Attributes:
        global_min: Global minimum value for time course plot - 
            computed from all time courses. Default is None
        global_max: Global maximum value for time course plot - 
            computed from all time courses. Default is None
        default_global_min: Default global minimum value for time course plot. 
            Default is -1.0
        default_global_max: Default global maximum value for time course plot.
             Default is 1.0
        grid_on: Whether grid is enabled for time course plot. Default is True
        hover_text_on: Whether hover text is enabled for time course plot. 
            Default is True
        time_marker_on: Whether time marker is enabled for time course plot. 
            Default is True
        global_convolution: Whether global convolution is enabled for time course plot. 
            Default is True
        shift_unit: Constant shift unit applied to time course data on constant shift changes. 
            Default is 1.0. This value is updated when global min and max are updated.
        shift_update_ratio: Ratio of shift unit updates. 
            Default is 10.
        scale_unit: Scale unit applied to time course data on scale change. 
            Default is 0.1. This value is updated when global min and max are updated.
        scale_update_granularity: Granularity of scale unit updates. 
            Default is 10.
    """
    global_min: Optional[float] = None
    global_max: Optional[float] = None
    default_global_min: float = -1.0
    default_global_max: float = 1.0
    grid_on: bool = True
    global_convolution: bool = True
    hover_text_on: bool = True
    time_marker_on: bool = True
    shift_unit: float = 1.0
    shift_update_ratio: int = 10
    scale_unit: float = 0.1
    scale_update_granularity: int = 10

    def to_dict(self) -> Dict[str, float]:
        """Convert to dictionary."""
        return {
            'global_min': self.global_min,
            'global_max': self.global_max,
            'grid_on': self.grid_on,
            'hover_text_on': self.hover_text_on,
            'time_marker_on': self.time_marker_on,
            'global_convolution': self.global_convolution,
            'shift_unit': self.shift_unit,
            'shift_update_ratio': self.shift_update_ratio,
            'scale_unit': self.scale_unit,
            'scale_update_granularity': self.scale_update_granularity
        }
    
    def update_from_dict(self, data: Dict[str, any]) -> None:
        """Update from dictionary."""
        for key, value in data.items():
            setattr(self, key, value)

@dataclass
class TimeCoursePlotOptions:
    """Time course plot options.
    
    Attributes:
        label: Label of the time course. Default is None
        visibility: whether the time course is visible in the plot. Default is True.
        color: Color of the time course. Default is RED
        width: Width of the time course. Default is 2.0
        constant: History of constant shifts to the time course handled by SignalShifter.
        scale: History of scale changes to the time course handled by SignalScaler.
        preprocess_constant: History of constant shifts to the preprocessed time course 
            handled by SignalShifter.
        preprocess_scale: History of scale changes to the preprocessed time course 
            handled by SignalScaler.
        opacity: Opacity of the time course. Default is 1.0
        mode: Mode of the time course. Default is 'lines'
    """
    label: str = None
    visibility: bool = True
    color: TimeCourseColor = TimeCourseColor.RED
    width: float = 2.0
    constant: SignalShifter = field(default_factory=SignalShifter)
    scale: SignalScaler = field(default_factory=SignalScaler)
    preprocess_constant: SignalShifter = field(default_factory=SignalShifter)
    preprocess_scale: SignalScaler = field(default_factory=SignalScaler)
    opacity: float = 1.0
    mode: Literal['lines', 'markers', 'lines+markers'] = 'lines'

    @classmethod
    def with_color(cls, color: TimeCourseColor, **kwargs):
        """Create TimeCoursePlotOptions with a specific color."""
        return cls(color=color, **kwargs)

    @classmethod
    def with_next_color(cls, used_colors: set[TimeCourseColor], **kwargs):
        """Create TimeCoursePlotOptions with the next available color.
        
        Args:
            used_colors: Set of already used TimeCourseColor values
            **kwargs: Additional arguments to pass to constructor
            
        Returns:
            TimeCoursePlotOptions with first unused color from TimeCourseColor enum
        """
        available_colors = set(TimeCourseColor) - used_colors
        if not available_colors:
            # If all colors used, start over with first color
            next_color = list(TimeCourseColor)[0]
        else:
            next_color = list(available_colors)[0]
        return cls(color=next_color, **kwargs)
    
    def clear_preprocess_history(self) -> None:
        """Clear preprocess history."""
        self.preprocess_constant.clear_history()
        self.preprocess_scale.clear_history()

    def to_dict(self) -> Dict[str, float]:
        """Convert to dictionary."""
        return {
            'label': self.label,
            'visibility': self.visibility,
            'color': self.color.value,
            'width': self.width,
            'constant': self.constant.shift_history,
            'scale': self.scale.scale_history,
            'preprocess_constant': self.preprocess_constant.shift_history,
            'preprocess_scale': self.preprocess_scale.scale_history,
            'opacity': self.opacity,
            'mode': self.mode
        }

    def update_from_dict(self, data: Dict[str, any]) -> None:
        """Update from dictionary."""
        for key, value in data.items():
            # handle enum values
            if key == 'color':
                setattr(self, key, TimeCourseColor(value))
            # handle SignalScaler and SignalShifter
            elif key == 'constant':
                self.constant.set_history(value)
            elif key == 'scale':
                self.scale.set_history(value)
            elif key == 'preprocess_constant':
                self.preprocess_constant.set_history(value)
            elif key == 'preprocess_scale':
                self.preprocess_scale.set_history(value)
            else:
                setattr(self, key, value)

@dataclass
class TimeMarkerPlotOptions:
    """Time marker plot options."""
    opacity: float = 0.5
    width: float = 1.0
    shape: Literal['solid', 'dash', 'dot', 'longdash', 'dashdot'] = 'solid'
    color: TimeCourseColor = TimeCourseColor.GREY

    def to_dict(self) -> Dict[str, float]:
        """Convert to dictionary."""
        return {
            'opacity': self.opacity,
            'width': self.width,
            'shape': self.shape,
            'color': self.color.value
        }
    
    def update_from_dict(self, data: Dict[str, float]) -> None:
        """Update from dictionary."""
        for key, value in data.items():
            # handle enum values
            if key == 'color':
                setattr(self, key, TimeCourseColor(value))
            else:
                setattr(self, key, value)


@dataclass
class TaskDesignPlotOptions:
    """Task design plot options.
    
    Attributes:
        label: Label of the task design. Default is None
        convolution: Whether convolution of task design with hrf is enabled, 
            otherwise block structure is plotted. Default is hrf.
        scale: History of scale changes to the task design handled by SignalScaler.
        constant: History of constant shifts to the task design handled by SignalShifter.
        color: Color of the task design. Default is RED
        width: Width of the task design. Default is 2.0
        opacity: Opacity of the task design. Default is 1.0
        mode: Mode of the task design. Default is 'lines'
    """
    label: str = None
    convolution: Literal['hrf', 'block'] = 'hrf'
    scale: SignalScaler = field(default_factory=SignalScaler)
    constant: SignalShifter = field(default_factory=SignalShifter)
    color: TimeCourseColor = TimeCourseColor.RED
    width: float = 2.0
    opacity: float = 1.0
    mode: Literal['lines', 'markers', 'lines+markers'] = 'lines'

    @classmethod
    def with_color(cls, color: TimeCourseColor, **kwargs):
        """Create TaskDesignPlotOptions with a specific color."""
        return cls(color=color, **kwargs)

    @classmethod
    def with_next_color(cls, used_colors: set[TimeCourseColor], **kwargs):
        """Create TaskDesignPlotOptions with the next available color.
        
        Args:
            used_colors: Set of already used TimeCourseColor values
            **kwargs: Additional arguments to pass to constructor
            
        Returns:
            TimeCoursePlotOptions with first unused color from TimeCourseColor enum
        """
        available_colors = set(TimeCourseColor) - used_colors
        if not available_colors:
            # If all colors used, start over with first color
            next_color = list(TimeCourseColor)[0]
        else:
            next_color = list(available_colors)[0]
        return cls(color=next_color, **kwargs)
    
    def to_dict(self) -> Dict[str, float]:
        """Convert to dictionary."""
        return {
            'label': self.label,
            'convolution': self.convolution,
            'scale': self.scale.scale_history,
            'constant': self.constant.shift_history,
            'color': self.color.value,
            'width': self.width,
            'opacity': self.opacity,
            'mode': self.mode
        }

    def update_from_dict(self, data: Dict[str, float]) -> None:
        """Update from dictionary."""
        for key, value in data.items():
            # handle enum values
            if key == 'color':
                setattr(self, key, TimeCourseColor(value))
            # handle SignalScaler and SignalShifter
            elif key == 'scale':
                self.scale.set_history(value)
            elif key == 'constant':
                self.constant.set_history(value)
            else:
                setattr(self, key, value)



import numpy as np


# index array with range (w/ NaN padding for out-of-bound indices )
def extract_range(array, center, left_edge, right_edge):
    """
    Extract a range of rows from an array with NaN padding for out-of-bound indices

    Parameters
    ----------
    array : np.ndarray
        array to extract range from
    center : int
        center of the range
    left_edge : int
        left edge of the range
    right_edge : int
        right edge of the range

    Returns
    -------
    padded_range : np.ndarray
        array with NaN padding for out-of-bound indices
    """
    num_rows, num_cols = array.shape
    range_size = right_edge - left_edge + 1
    # Create a NaN-filled placeholder for the range
    padded_range = np.full((range_size, num_cols), np.nan)

    # Calculate valid row indices
    start = max(0, center + left_edge)  # Valid start index in the array
    end = min(num_rows, center + right_edge + 1)  # Valid end index in the array
    insert_start = max(0, -1 * (center + left_edge))  # Where to insert in padded_range
    insert_end = insert_start + (end - start)  # End position in padded_range

    # Insert valid rows into the padded range
    padded_range[insert_start:insert_end, :] = array[start:end, :]
    return padded_range


def get_lag_mat(x: np.ndarray, lags: list) -> np.ndarray:
    """
    Create array of time-lagged copies of the time course. Modified
    for negative lags from:
    https://github.com/ulf1/lagmat

    Parameters
    ----------
    x : np.ndarray
        time course (n_timepoints)
    lags : list
        lags (n_lags)

    Returns
    -------
    x_lag : np.ndarray
        time-lagged copies of the time course (n_timepoints, n_lags * n_timepoints)
    """
    n_rows, n_cols = x.shape
    n_lags = len(lags)
    # return if no lags
    if n_lags < 1:
        return x
    # allocate memory
    x_lag = np.zeros(
        shape=(n_rows, n_cols * n_lags),
        order='F', dtype=x.dtype
    )
    # Copy lagged columns of X into X_lag
    for i, l in enumerate(lags):
        # target columns of X_lag
        j = i * n_cols
        k = j + n_cols  # (i+1) * ncols
        
        # Skip if lag is out of bounds (lag >= array length)
        if abs(l) >= n_rows:
            continue
            
        # number rows of X
        nl = n_rows - abs(l)
        
        # Copy
        if l >= 0:
            x_lag[l:, j:k] = x[:nl, :]
        else:
            x_lag[:l, j:k] = x[-nl:, :]
    return x_lag




"""
Routes for preprocessing of FMRI and timecourse data

Routes:
    GET_PREPROCESSED_FMRI: Get preprocessed FMRI data
    GET_PREPROCESSED_TIMECOURSE: Get preprocessed timecourse data
    RESET_FMRI_PREPROCESS: Reset fmri preprocessing
    RESET_TIMECOURSE_PREPROCESS: Reset timecourse preprocessing
"""
import json

from flask import Blueprint, request, make_response

from findviz.logger_config import setup_logger
from findviz.routes.shared import data_manager
from findviz.routes.utils import convert_value, Routes, handle_route_errors
from findviz.viz.preprocess.fmri import preprocess_fmri, PreprocessFMRIInputs
from findviz.viz.preprocess.timecourse import preprocess_timecourse, PreprocessTimecourseInputs
from findviz.viz.preprocess.input import (
    FMRIPreprocessInputValidator, 
    TimecoursePreprocessInputValidator
)
from findviz.viz.exception import PreprocessInputError, NiftiMaskError

logger = setup_logger(__name__)
preprocess_bp = Blueprint('preprocess', __name__)

@preprocess_bp.route(Routes.GET_PREPROCESSED_FMRI.value, methods=['POST'])
@handle_route_errors(
    error_msg='Unknown error in preprocess FMRI request',
    log_msg='FMRI preprocessing successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_PREPROCESSED_FMRI,
    route_parameters=list(PreprocessFMRIInputs.__annotations__.keys()),
    custom_exceptions=[NiftiMaskError, PreprocessInputError]
)
def get_preprocessed_fmri() -> dict:
    """Get preprocessed FMRI data"""
    if data_manager.ctx.fmri_preprocessed:
        logger.info("FMRI data already preprocessed, clearing it")
        data_manager.ctx.clear_fmri_preprocessed()

    params = {key: convert_value(value) for key, value in request.form.items()}
    inputs = PreprocessFMRIInputs(**params)
    logger.info(f"Preprocessing FMRI data with inputs: {inputs}")

    # Validate inputs
    fmri_input_validator = FMRIPreprocessInputValidator(data_manager.ctx.fmri_file_type)
    fmri_input_validator.validate_preprocess_input(inputs)

    # get fmri data
    viewer_data = data_manager.ctx.get_viewer_data(
        fmri_data=True,
        time_course_data=False,
        task_data=False,
    )

    # preprocess fmri data
    if data_manager.ctx.fmri_file_type == 'nifti':
        func_proc = preprocess_fmri(
            file_type=data_manager.ctx.fmri_file_type,
            inputs=inputs,
            func_img=viewer_data['func_img'],
            mask_img=viewer_data['mask_img'],
        )
        data_manager.ctx.store_fmri_preprocessed({'func_img': func_proc})
    else:
        func_proc = preprocess_fmri(
            file_type=data_manager.ctx.fmri_file_type,
            inputs=inputs,
            left_func_img=viewer_data['left_func_img'],
            right_func_img=viewer_data['right_func_img'],
        )
        data_manager.ctx.store_fmri_preprocessed({
            'left_func_img': func_proc[0], 
            'right_func_img': func_proc[1]
        })
    
    logger.info(f"Preprocessed FMRI data successfully")
    
    return {'status': 'success'}


@preprocess_bp.route(Routes.GET_PREPROCESSED_TIMECOURSE.value, methods=['POST'])
@handle_route_errors(
    error_msg='Unknown error in preprocess timecourse request',
    log_msg='Timecourse preprocessing successful',
    route=Routes.GET_PREPROCESSED_TIMECOURSE,
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route_parameters=list(PreprocessTimecourseInputs.__annotations__.keys()),
    custom_exceptions=[PreprocessInputError]
)
def get_preprocessed_timecourse() -> dict:
    """Get preprocessed timecourse data"""
    ts_labels = json.loads(request.form['ts_labels'])
    params = {
        key: convert_value(value) for key, value in request.form.items()
        if key != 'ts_labels'
    }
    inputs = PreprocessTimecourseInputs(**params)
    logger.info(f"Preprocessing timecourse data with inputs: {inputs}")

    # Validate inputs
    timecourse_input_validator = TimecoursePreprocessInputValidator()
    timecourse_input_validator.validate_preprocess_input(
        inputs, ts_labels
    )

    # get timecourse data
    viewer_data = data_manager.ctx.get_viewer_data(
        fmri_data=False,
        time_course_data=True,
        task_data=False,
    )

    # preprocess timecourse data
    ts_data = {}
    for ts_label in ts_labels:
        ts_proc = preprocess_timecourse(
            timecourse_data=viewer_data['ts'][ts_label],
            inputs=inputs
        )
        ts_data[ts_label] = ts_proc

    # store preprocessed timecourse data
    data_manager.ctx.store_timecourse_preprocessed(ts_data)
    return {'status': 'success'}


@preprocess_bp.route(Routes.RESET_FMRI_PREPROCESS.value, methods=['POST'])
@handle_route_errors(
    error_msg='Unknown error in reset FMRI preprocessing request',
    log_msg='FMRI preprocessing reset successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.RESET_FMRI_PREPROCESS
)
def reset_fmri_preprocess() -> dict:
    """Reset FMRI preprocessing"""
    data_manager.ctx.clear_fmri_preprocessed()
    return {'status': 'success'}


@preprocess_bp.route(Routes.RESET_TIMECOURSE_PREPROCESS.value, methods=['POST'])
@handle_route_errors(
    error_msg='Unknown error in reset timecourse preprocessing request',
    log_msg='Timecourse preprocessing reset successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.RESET_TIMECOURSE_PREPROCESS,
    route_parameters=['ts_labels'],
    custom_exceptions=[PreprocessInputError]
)
def reset_timecourse_preprocess() -> dict:
    """Reset timecourse preprocessing"""
    ts_labels = json.loads(request.form['ts_labels'])
    # check if no timecourse labels provided
    if len(ts_labels) == 0:
        raise PreprocessInputError(
            'No timecourses selected', 
            preprocess_method='reset'
        )
    # check if time course label is in preprocessed data
    for ts_label in ts_labels:
        if ts_label not in data_manager.ctx.ts_labels_preprocessed:
            raise PreprocessInputError(
                f'Timecourse {ts_label} is not preprocessed',
                preprocess_method='reset'
            )
    # clear timecourse preprocessed data
    data_manager.ctx.clear_timecourse_preprocessed(ts_labels)
    return {'status': 'success'}



"""
Utilities for handling gifti file uploads
"""
from enum import Enum
from pathlib import Path
from typing import Literal, Optional, Dict, Union, Tuple

import nibabel as nib

from flask import request
from werkzeug.datastructures import FileStorage

from findviz.viz import exception
from findviz.viz.io import utils
from findviz.viz.io import validate

# Type aliases
FilePath = Union[str, Path]
FileUploadDict = Dict[str, Optional[Union[FilePath, 'FileStorage']]]
GiftiDict = Dict[str, Optional[nib.GiftiImage]]

# Expected file upload inputs
class GiftiFiles(Enum):
    LEFT_FUNC = 'left_gii_func'
    RIGHT_FUNC = 'right_gii_func'
    LEFT_MESH = 'left_gii_mesh'
    RIGHT_MESH = 'right_gii_mesh'

# gifti form fields in upload modal
browser_fields: Dict[str, str] = {
    GiftiFiles.LEFT_FUNC.value: 'left-hemisphere-gifti-func',
    GiftiFiles.RIGHT_FUNC.value: 'right-hemisphere-gifti-func',
    GiftiFiles.LEFT_MESH.value: 'left-hemisphere-gifti-mesh',
    GiftiFiles.RIGHT_MESH.value: 'right-hemisphere-gifti-mesh'
}


class GiftiUpload:
    """
    Class for handling gifti file uploads.

    Attributes
    ----------
    method : Literal['cli', 'browser']
        The method used for file uploads
    """

    def __init__(
        self,
        method: Literal['cli', 'browser']
    ):
        if method not in ['cli', 'browser']:
            raise ValueError('unrecognized upload method')
        self.method = method

    def upload(
        self,
        fmri_files: Optional[Dict[str, Union[str, Path]]] = None,
    ) -> GiftiDict:
        """
        Get and validate gifti files uploaded from cli or browser.

        Parameters
        ----------
        fmri_files : Optional[Dict[str, FilePath]]
            Dictionary of Gifti file paths: 
            {'left_gii_func': path, 'right_gii_func': path, 'left_gii_mesh': path, 'right_gii_mesh': path}

        Returns
        -------
        GiftiDict
            Dictionary containing validated Nifti images

        Raises
        ------
        FileInputError
            If required files are missing or have invalid extensions
        FileUploadError
            If there are issues reading the files
        FileValidationError
            If files fail validation checks
        """
        if self.method == 'cli':
            file_uploads = fmri_files
        elif self.method == 'browser':
            file_uploads = self._get_browser_input()

        # check file uploads
        self._check_file_inputs(file_uploads)

        # store whether left and/or right files were uploaded
        if file_uploads[GiftiFiles.LEFT_FUNC.value] is None:
            self.left_input = False
        else:
            self.left_input = True

        if file_uploads[GiftiFiles.RIGHT_FUNC.value] is None:
            self.right_input = False
        else:
            self.right_input = True
        
        # Initialize output dictionary
        gifti_out = {
            GiftiFiles.LEFT_FUNC.value: None,
            GiftiFiles.RIGHT_FUNC.value: None,
            GiftiFiles.LEFT_MESH.value: None,
            GiftiFiles.RIGHT_MESH.value: None
        }

        # Process left hemisphere inputs
        if self.left_input:
            gii_left_func, gii_left_mesh = self._process_hemisphere(
                file_uploads[GiftiFiles.LEFT_FUNC.value],
                file_uploads[GiftiFiles.LEFT_MESH.value],
                'left'
            )
            gifti_out[GiftiFiles.LEFT_FUNC.value] = gii_left_func
            gifti_out[GiftiFiles.LEFT_MESH.value] = gii_left_mesh

        # Process right hemisphere inputs
        if self.right_input:
            gii_right_func, gii_right_mesh = self._process_hemisphere(
                file_uploads[GiftiFiles.RIGHT_FUNC.value],
                file_uploads[GiftiFiles.RIGHT_MESH.value],
                'right'
            )
            gifti_out[GiftiFiles.RIGHT_FUNC.value] = gii_right_func
            gifti_out[GiftiFiles.RIGHT_MESH.value] = gii_right_mesh
            
        # if left and right hemispheres passed, check length is consistent
        if self.left_input & self.right_input:
            if not validate.validate_gii_func_len(
                gifti_out[GiftiFiles.LEFT_FUNC.value], 
                gifti_out[GiftiFiles.RIGHT_FUNC.value]
            ):
                raise exception.FileValidationError(
                    "The left and right hemisphere func.gii files are "
                    "not the same length. Check files are consistent in length.",
                    validate.validate_gii_func_len.__name__,
                    exception.ExceptionFileTypes.GIFTI.value,
                    [
                        browser_fields[GiftiFiles.LEFT_FUNC.value], 
                        browser_fields[GiftiFiles.RIGHT_FUNC.value]
                    ]
                )
            
        return gifti_out

    def _check_file_inputs(self, file_inputs: FileUploadDict) -> None:
        """
        Check left and right hemisphere file uploads.

        Parameters
        ----------
        file_inputs : FileUploadDict
            Dictionary containing file uploads

        Raises
        ------
        FileValidationError
            If file inputs are incorrect or inconsistent
        """
        # get files from dict
        left_func = file_inputs[GiftiFiles.LEFT_FUNC.value]
        right_func = file_inputs[GiftiFiles.RIGHT_FUNC.value]
        left_mesh = file_inputs[GiftiFiles.LEFT_MESH.value]
        right_mesh = file_inputs[GiftiFiles.RIGHT_MESH.value]

        msg, missing, passed = validate.validate_gii_file_inputs(
            left_mesh, 
            right_mesh, 
            left_func, 
            right_func
        )
        # missing files to browser fields
        missing_dict = {
            'left_func': browser_fields[GiftiFiles.LEFT_FUNC.value],
            'right_func': browser_fields[GiftiFiles.RIGHT_FUNC.value],
            'left_mesh': browser_fields[GiftiFiles.LEFT_MESH.value],
            'right_mesh': browser_fields[GiftiFiles.RIGHT_MESH.value],
        }
        # raise file validation error if file inputs incorrect
        if not passed:
            missing_fields = [missing_dict[file] for file in missing]
            raise exception.FileValidationError(
                msg, exception.ExceptionFileTypes.GIFTI.value, self.method,
                missing_fields
            )

    def _process_hemisphere(
        self, 
        func_gii_file: FilePath,
        mesh_gii_file: FilePath, 
        hemisphere: Literal['left', 'right']
    ) -> Tuple[nib.GiftiImage, nib.GiftiImage]:
        """
        Process left or right hemisphere file uploads.

        Parameters
        ----------
        func_gii : nib.GiftiImage
            Functional gifti file
        mesh_gii : nib.GiftiImage
            Mesh gifti file
        hemisphere : Literal['left', 'right']
            Hemisphere to process
        
        Returns
        -------
        Tuple[nib.GiftiImage, nib.GiftiImage]
            Functional and mesh gifti files
        """
        # load and validate left functional file
        # get browser field for hemisphere
        if hemisphere == 'left':
            browser_field_func = browser_fields[GiftiFiles.LEFT_FUNC.value]
            browser_field_mesh = browser_fields[GiftiFiles.LEFT_MESH.value]
        else:
            browser_field_func = browser_fields[GiftiFiles.RIGHT_FUNC.value]
            browser_field_mesh = browser_fields[GiftiFiles.RIGHT_MESH.value]

        # check left functional file extension
        if not validate.validate_gii_func_ext(
            utils.get_filename(func_gii_file)
        ):
            raise exception.FileInputError(
                'File extension is not recognized. Please upload '
                f'a gifti file with a .func.gii extension for {hemisphere} '
                'hemisphere functional file',
                exception.ExceptionFileTypes.GIFTI.value, self.method,
                [browser_field_func]
            )
        try:
            gii_func = read_gii(func_gii_file, self.method)
        except Exception as e:
            raise exception.FileUploadError(
                f'Error in reading {hemisphere} functional gifti file',
                exception.ExceptionFileTypes.GIFTI.value, self.method,
                [browser_field_func]
            ) from e
        
        if not validate.validate_gii_func(gii_func):
            raise exception.FileValidationError(
                f"The {hemisphere} hemisphere func.gii file should have have 1d-array "
                "per data array (timepoint). Check format.",
                validate.validate_gii_func.__name__,
                exception.ExceptionFileTypes.GIFTI.value,
                [browser_field_func]
            )

        # load and validate left mesh
        # check left mesh file extension
        if not validate.validate_gii_mesh_ext(
            utils.get_filename(mesh_gii_file)
        ):
            raise exception.FileInputError(
                'File extension is not recognized. Please upload '
                f'a gifti file with a .surf.gii extension for {hemisphere} '
                'hemisphere mesh (geometry) file',
                exception.ExceptionFileTypes.GIFTI.value, self.method,
                [browser_field_mesh]
            )
        try:
            gii_mesh = read_gii(mesh_gii_file, self.method)
        except Exception as e:
            raise exception.FileUploadError(
                f'Error in reading {hemisphere} hemisphere mesh (geometry) gifti file',
                exception.ExceptionFileTypes.GIFTI.value, self.method,
                [browser_field_mesh]
            ) from e
        
        if not validate.validate_gii_mesh(gii_mesh):
            raise exception.FileValidationError(
                f"The {hemisphere} hemisphere mesh (geometry) file should only contain "
                "two data arrays, corresponding to coordinate and face arrays."
                    " Check format.",
                validate.validate_gii_mesh.__name__,
                exception.ExceptionFileTypes.GIFTI.value,
                [browser_field_mesh]
            )
        
        # check length of functional and mesh files
        if not validate.validate_gii_func_mesh_len(gii_func, gii_mesh):
            raise exception.FileValidationError(
                f"The {hemisphere} hemisphere func.gii file and mesh (geometry) file "
                f"are not the same length - {len(gii_func.darrays[0].data)} != "
                f"{len(gii_mesh.darrays[0].data)}. Check files are consistent "
                "in length.",
                validate.validate_gii_func_mesh_len.__name__,
                exception.ExceptionFileTypes.GIFTI.value,
                [
                    browser_field_func, 
                    browser_field_mesh
                ]
            )
        return gii_func, gii_mesh

    @staticmethod
    def _get_browser_input() -> FileUploadDict:
        """
        Get file uploads from browser request.

        Returns
        -------
        FileUploadDict
            Dictionary containing file upload objects
        """
        left_func_file = request.files.get(GiftiFiles.LEFT_FUNC.value)
        right_func_file = request.files.get(GiftiFiles.RIGHT_FUNC.value)
        left_mesh_file = request.files.get(GiftiFiles.LEFT_MESH.value)
        right_mesh_file = request.files.get(GiftiFiles.RIGHT_MESH.value)

        file_upload = {
            GiftiFiles.LEFT_FUNC.value: left_func_file,
            GiftiFiles.RIGHT_FUNC.value: right_func_file,
            GiftiFiles.LEFT_MESH.value: left_mesh_file,
            GiftiFiles.RIGHT_MESH.value: right_mesh_file
        }
        return file_upload


def read_gii(
    file: Union[FilePath, FileStorage],
    method: Literal['cli', 'browser']
) -> nib.GiftiImage:
    """
    Read gifti file based on method of upload.

    Parameters
    ----------
    file : Union[FilePath, FileStorage]
        Either a gifti file from the browser or full path to file from CLI
    method : Literal['cli', 'browser']
        Whether the file was uploaded through browser or CLI

    Returns
    -------
    nib.GiftiImage
        Loaded nibabel image

    Raises
    ------
    Exception
        If there are issues reading the file
    """
    if method =='cli':
        try:
            gifti_img = nib.load(file)
        # raise generic exception to be handled higher in stack
        except Exception as e:
            raise e
    elif method == 'browser':
        try:
            file_bytes = file.read()
            gifti_img = nib.GiftiImage.from_bytes(file_bytes)
        except Exception as e:
            raise e
        
    return gifti_img



"""
Utilities for loading and validating file uploads
"""
from enum import Enum
from typing import Literal, List, Dict, Optional, Union, TypedDict
from pathlib import Path

from findviz.logger_config import setup_logger
from findviz.viz import exception
from findviz.viz.io import cifti
from findviz.viz.io import gifti
from findviz.viz.io import nifti
from findviz.viz.io import timecourse

# Set up a logger for the app
logger = setup_logger(__name__)

# HTML Div IDs for upload modal 
browser_fields = {
    'nifti': nifti.browser_fields,
    'gifti': gifti.browser_fields,
    'ts': timecourse.browser_fields
}


class FileUpload:
    """
    Class for handling file uploads through browser or CLI.
    
    File inputs are not stored in class.

    Attributes
    ----------
    fmri_file_type : Literal['nifti', 'gifti', 'cifti']
        The fMRI file type
    ts_status : bool
        Whether time course files were provided
    task_status : bool
        Whether task design files were provided
    method : Literal['cli', 'browser']
        The method used for file uploads
    upload_status : Union[bool, Dict[str, bool]]
        State variable tracking upload status
    fmri_uploader : Union[NiftiUpload, GiftiUpload]
        Uploader for fMRI files
    fmri_file_labels : Union[NiftiFiles, GiftiFiles]
        File labels for fMRI files
    ts_uploader : Optional[TimeCourseUpload]
        Uploader for time series files
    ts_files : Optional[SingleTimeCourseFiles]
        File labels for time series
    task_uploader : Optional[TaskDesignUpload]
        Uploader for task design files
    task_files : Optional[TaskDesignFiles]
        File labels for task design
    """

    def __init__(
        self,
        fmri_file_type: Literal['nifti', 'gifti', 'cifti'],
        ts_status: bool,
        task_status: bool,
        method: Literal['cli', 'browser']
    ):
        """
        Initialize FileUpload instance.

        Parameters
        ----------
        fmri_file_type : Literal['nifti', 'gifti', 'cifti']
            Type of fMRI files to process
        ts_status : bool
            Whether time series files will be uploaded
        task_status : bool
            Whether task design files will be uploaded
        method : Literal['cli', 'browser']
            Method of file upload (command line or browser)

        Raises
        ------
        ValueError
            If fmri_file_type is not 'nifti', 'gifti', or 'cifti'
        """
        self.fmri_file_type = fmri_file_type
        self.ts_status = ts_status
        self.task_status = task_status
        self.method = method
        # state variable for whether files have been uploaded
        self.upload_status = False
        # get fmri file uploaders and file labels
        if fmri_file_type == 'gifti':
            self.fmri_uploader = gifti.GiftiUpload(method)
            self.fmri_file_labels = gifti.GiftiFiles
        elif fmri_file_type == 'nifti':
            self.fmri_uploader = nifti.NiftiUpload(method)
            self.fmri_file_labels = nifti.NiftiFiles
        elif fmri_file_type == 'cifti':
            self.fmri_uploader = cifti.CiftiUpload(method)
            self.fmri_file_labels = cifti.CiftiFiles
        else:
            raise ValueError('unrecognized fmri file type [nifti, gifti, cifti]')

        # get time course file uploaders and file labels
        if self.ts_status:
            self.ts_uploader = timecourse.TimeCourseUpload(method)
            self.ts_files = timecourse.SingleTimeCourseFiles
        else:   
            self.ts_uploader = None
            self.ts_files = None

        # get task design file uploaders and file labels
        if self.task_status:
            self.task_uploader = timecourse.TaskDesignUpload(method)
            self.task_files = timecourse.TaskDesignFiles
        else:
            self.task_uploader = None
            self.task_files = None

    def check_files(
        self, 
        key: Literal['gifti', 'nifti', 'cifti', 'ts', 'task']
    ) -> bool:
        """
        Check whether file type was uploaded by user.

        Parameters
        ----------
        key : Literal['gifti', 'nifti', 'cifti', 'ts', 'task']
            The file type to check

        Returns
        -------
        bool
            Whether files of the specified type were uploaded

        Raises
        ------
        AttributeError
            If no files have been uploaded yet
        KeyError
            If key is not a recognized file label
        """
        if not self.upload_status:
            raise AttributeError('No files have been uploaded!')

        try:
            return self.upload_status[key]
        except KeyError as e:
            raise Exception(f'{key} is not a recognized file label') from e

    def hemisphere(self) -> Literal['left', 'right', 'both']:
        """
        Check which hemispheres were uploaded in GIFTI file upload.

        Returns
        -------
        Literal['left', 'right', 'both']
            Which hemisphere(s) were uploaded

        Raises
        ------
        AttributeError
            If no files have been uploaded yet or if using NIFTI files
        """
        if not self.upload_status:
            raise AttributeError('No files have been uploaded!')

        if self.fmri_file_type == 'nifti':
            raise AttributeError(
                'hemisphere function not supported for nifti file inputs'
            )

        if self.fmri_file_type == 'gifti':
            if self.fmri_uploader.left_input & self.fmri_uploader.right_input:
                return 'both'
            elif self.fmri_uploader.left_input:
                return 'left'
            elif self.fmri_uploader.right_input:
                return 'right'

    def upload(
        self,
        fmri_files: Optional[Dict[str, Union[str, Path]]] = None,
        ts_files: Optional[List[Union[str, Path]]] = None,
        ts_labels: Optional[List[str]] = None,
        ts_headers: Optional[List[bool]] = None,
        task_file: Optional[Union[str, Path]] = None,
        tr: Optional[float] = None,
        slicetime_ref: Optional[float] = None
    ) -> Dict:
        """
        Upload and validate files from either browser or CLI.

        Parameters
        ----------
        fmri_files : Dict[str, Union[str, Path]], optional
            Dictionary of FMRI file paths. For NIFTI: {'nii_func': path, 'nii_anat': path, 'nii_mask': path}
            For GIFTI: {'left_gii_func': path, 'right_gii_func': path, 'left_gii_mesh': path, 'right_gii_mesh': path}
            For CIFTI: {'dtseries': path, 'left_gii_mesh': path, 'right_gii_mesh': path}
        ts_files : List[Union[str, Path]], optional
            List of paths to time series files
        ts_labels : List[str], optional
            List of labels for time series files
        ts_headers : List[bool], optional
            List of boolean flags indicating if time series files have headers
        task_file : Union[str, Path], optional
            Path to task design file
        tr : float, optional
            Repetition time value
        slicetime_ref : float, optional
            Slice timing reference value (0-1)
            
        Returns
        -------
        FileOutput
            Dictionary containing processed and validated file data
        
        Raises
        ------
        FileInputError
            If required files are missing
        FileUploadError
            If there are issues uploading files
        FileValidationError
            If files fail validation
        """
        # init output dictionary
        file_out = {
            "gifti": None,
            "nifti": None,
            "ts": None,
            "task": None
        }
        
        # init upload status dictionary
        self.upload_status: Dict[str, bool] = {
            "gifti": False,
            "nifti": False,
            "ts": False,
            "task": False
        }
        
        # get files from CLI or browser
        try:
            # upload fmri files
            fmri_files = self.fmri_uploader.upload(fmri_files)
            # store fmri files
            if self.fmri_file_type == 'gifti' or self.fmri_file_type == 'cifti':
                file_out['gifti'] = fmri_files
                self.upload_status['gifti'] = True
            elif self.fmri_file_type == 'nifti':
                file_out['nifti'] = fmri_files
                self.upload_status['nifti'] = True
            
            # get # of volumes from functional MRI
            if self.fmri_file_type == 'nifti':
                fmri_len = file_out['nifti'][nifti.NiftiFiles.FUNC.value].shape[-1]
            elif self.fmri_file_type == 'gifti' or self.fmri_file_type == 'cifti':
                if self.fmri_uploader.left_input:
                    gifti_out = file_out['gifti'][gifti.GiftiFiles.LEFT_FUNC.value].darrays
                else:
                    gifti_out = file_out['gifti'][gifti.GiftiFiles.RIGHT_FUNC.value].darrays
                fmri_len = len(gifti_out)

            # get time series files
            if self.ts_status:
                ts_files = self.ts_uploader.upload(
                    fmri_len=fmri_len,
                    ts_files=ts_files,
                    ts_labels=ts_labels,
                    ts_headers=ts_headers
                )
                file_out['ts'] = ts_files
                self.upload_status['ts'] = True

            # get task design files
            if self.task_status:
                task_files = self.task_uploader.upload(
                    fmri_len=fmri_len,
                    task_file=task_file,
                    tr=tr,
                    slicetime_ref=slicetime_ref
                )
                file_out['task'] = task_files
                self.upload_status['task'] = True

        except (
            exception.FileInputError,
            exception.FileUploadError,
            exception.FileValidationError
        ) as e:
            # propagate file error returned from upload function
            raise e

        # update state variable
        self.upload_status = True

        return file_out

        


"""
GIFTI data handling
"""

from typing import Tuple, Optional, List, Literal, TypedDict

import nibabel as nib
import numpy as np

from findviz.routes.utils import sanitize_array_for_json


class GiftiTimePointData(TypedDict):
    left_hemisphere: List[float]
    right_hemisphere: List[float]

def get_gifti_data(
    time_point: int,
    left_func_img: Optional[nib.GiftiImage],
    right_func_img: Optional[nib.GiftiImage],
    threshold_min: float = 0,
    threshold_max: float = 0,
    threshold_min_orig: float = 0,
    threshold_max_orig: float = 0
) -> GiftiTimePointData:
    """Get functional data for a specific timepoint from left and right hemisphere Gifti images.
    
    Parameters:
    -----------
        time_point : int
            The timepoint to extract data from
        left_func_img : nib.GiftiImage
            Gifti image containing left hemisphere functional data
        right_func_img : nib.GiftiImage
            Gifti image containing right hemisphere functional data
        threshold_min : float
            Minimum threshold value
        threshold_max : float
            Maximum threshold value
        threshold_min_orig : float
            Original minimum threshold value
        threshold_max_orig : float
            Original maximum threshold value
    Returns:
    --------
        Dictionary containing:
            - Left hemisphere functional data array for the timepoint (None if no left data)
            - Right hemisphere functional data array for the timepoint (None if no right data)
    """
    func_data_left = None
    func_data_right = None

    # Handle left hemisphere Gifti file
    if left_func_img is not None:
        func_data_left = left_func_img.darrays[time_point].data.copy()
        # threshold data if threshold_min or threshold_max have been changed
        if (threshold_min != threshold_min_orig) or (threshold_max != threshold_max_orig):
            func_data_left = threshold_gifti_data(
                func_data_left, threshold_min, threshold_max
            )
        # sanitize data
        func_data_left = sanitize_array_for_json(func_data_left)

    # Handle right hemisphere Gifti file
    if right_func_img is not None:
        func_data_right = right_func_img.darrays[time_point].data.copy()
        # threshold data if threshold_min or threshold_max have been changed
        if (threshold_min != threshold_min_orig) or (threshold_max != threshold_max_orig):
            func_data_right = threshold_gifti_data(
                func_data_right, threshold_min, threshold_max
            )
        # sanitize data
        func_data_right = sanitize_array_for_json(func_data_right)

    gifti_data = {
        'left_hemisphere': func_data_left,
        'right_hemisphere': func_data_right
    }

    return gifti_data


def get_timecourse_gifti(
    left_func_img: Optional[nib.GiftiImage],
    right_func_img: Optional[nib.GiftiImage],
    vertex_index: int,
    hemisphere: Literal['left', 'right'],
) -> Tuple[List[float], str]:
    """Get functional time course and label for a specific vertex 
    from left or right hemisphere Gifti images
    
    Parameters:
    ------------  
        left_func_img : nib.GiftiImage
            Gifti image containing left hemisphere functional data
        right_func_img : nib.GiftiImage
            Gifti image containing right hemisphere functional data
        vertex_index : int
            The vertex index to extract data from
        hemisphere : Literal['left', 'right']
            The hemisphere to extract data from
    
    Returns:
    --------
        Tuple containing:
            - Functional time course for vertex
            - Label for functional time course
    """
    if hemisphere == 'left':
        gifti_img = left_func_img
    elif hemisphere == 'right':
        gifti_img = right_func_img

    # Extract the vertex's time course
    time_course = [d.data[vertex_index] for d in gifti_img.darrays]
    time_course = [t.item() for t in time_course]

    # create time course label
    time_course_label = f'Vertex: {vertex_index} ({hemisphere})'

    return time_course, time_course_label


def threshold_gifti_data(
    gifti_data: np.ndarray,
    threshold_min: float,
    threshold_max: float,
) -> np.ndarray:
    """Threshold a list of functional data
    
    Parameters:
    -----------
        gifti_data: np.ndarray
            List of functional data
        threshold_min: float
            Minimum threshold value
        threshold_max: float
            Maximum threshold value
    
    Returns:
    --------
        np.ndarray of thresholded functional data
    """
    # create mask
    mask = (gifti_data >= threshold_min) & (gifti_data <= threshold_max)
    # apply mask
    gifti_data[mask] = np.nan
    return gifti_data



"""
Backend Data Management with DataManager singleton pattern.

This module provides a centralized data management system for the FIND viewer,
handling multiple visualization states (both NIFTI and GIFTI). 
It implements a singleton pattern to ensure consistent state across the application.

Classes:
    DataManager: Singleton manager for visualization state
"""

from typing import Dict, Optional, ClassVar, List

from findviz.logger_config import setup_logger
from findviz.viz.viewer.context import VisualizationContext
from findviz.viz.viewer.state.state_file import StateFile
from findviz.viz import exception
logger = setup_logger(__name__)


class DataManager:
    """Singleton manager for visualization state with support for multiple contexts.
    
    This class implements the singleton pattern to maintain a single source of truth
    for the visualization state across the entire application. It handles both NIFTI
    and GIFTI data types and their associated metadata.
    
    Attributes:
        _instance (ClassVar[Optional['DataManager']]): Singleton instance
        _contexts (Dict[str, VisualizationContext]): Dictionary mapping context IDs to contexts
        _active_context_id (str): ID of the currently active context
    
    Methods:
        __new__: Create or return the singleton instance
        active_context: Get the currently active visualization context
        create_analysis_context: Create a new context for analysis results
        get_context: Get a context by its ID
        get_context_ids: Get all available context IDs
        load: Load a scene file
        get_active_context_id: Get the ID of the currently active context
        save: save the current scene
        switch_context: Switch the active context to the specified ID
    """
    _instance: ClassVar[Optional['DataManager']] = None
    _contexts: Dict[str, VisualizationContext]
    _active_context_id: str
    
    def __new__(cls) -> 'DataManager':
        """Create or return the singleton instance."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            # Initialize instance attributes 
            cls._instance._contexts = {}  # Dictionary to store multiple visualization contexts
            cls._instance._active_context_id = "main"  # Default context is "main"
            cls._instance._contexts["main"] = VisualizationContext("main")
            logger.info("Data manager initialized")
        return cls._instance
    
    @property
    def ctx(self) -> VisualizationContext:
        """Short alias for the currently active visualization context."""
        return self._contexts[self._active_context_id]
    
    @property
    def _state(self) -> VisualizationContext:
        """Private property to get the currently active visualization state."""
        return self._contexts[self._active_context_id]

    
    def create_analysis_context(self, analysis_name: str) -> str:
        """Create a new context for analysis results.
        
        Arguments:
            analysis_name: Name of the analysis for context identification
            
        Returns:
            str: The ID of the newly created context
        """
        context_id = analysis_name
        self._contexts[context_id] = VisualizationContext(context_id)
        logger.info(f"Created new analysis context: {context_id}")
        return context_id
    
    def get_context(self, context_id: str) -> VisualizationContext:
        """Get a context by its ID.
        
        Arguments:
            context_id: The ID of the context to get

        Returns:
            VisualizationContext: The context with the specified ID
            
        Raises:
            ValueError: If the context does not exist
        """
        if context_id not in self._contexts:
            raise ValueError(f"Context {context_id} does not exist")
        return self._contexts[context_id]

    def get_context_ids(self) -> List[str]:
        """Get all available context IDs.
        
        Returns:
            List[str]: List of context IDs
        """
        return list(self._contexts.keys())
    
    def get_active_context_id(self) -> str:
        """Get the ID of the currently active context.
        
        Returns:
            str: ID of the active context
        """
        return self._active_context_id
    
    def load(self, scene_file_data: bytes) -> None:
        """Load a scene file and replace current context.
        
        Arguments:
            scene_file_data: The data to load
        """
        try:
            # Deserialize context from bytes
            context = StateFile.deserialize_from_bytes(scene_file_data)

            # Replace the main context with the loaded one
            self.switch_context("main")
            self._contexts["main"] = context
            logger.info(f"Loaded context from file with ID: {context.context_id}")
        # handle version incompatibility error higher up
        except exception.FVStateVersionIncompatibleError as e:
            raise
        except ValueError as e:
            logger.error(f"Error loading state file: {str(e)}")
            raise
        except Exception as e:
            logger.error(f"Unexpected error during state loading: {str(e)}")
            raise ValueError(f"Failed to load state: {str(e)}")

    def save(self, file_name: str = "scene.fvstate") -> bytes:
        """Save the current scene to a .fvstate file and return the serialized data.
        
        Arguments:
            file_name: Name of the file to save to
            
        Returns:
            bytes: Serialized context data that can be served as a download
        """        
        # Get current context
        context = self.get_context(self._active_context_id)
        
        # Serialize context using our custom format
        serialized_data = StateFile.serialize_to_bytes(context)
        logger.info(f"Prepared context {self._active_context_id} for download as {file_name}")
        
        return serialized_data

    def switch_context(self, context_id: str) -> None:
        """Switch the active context to the specified ID.
        
        Arguments:
            context_id: The ID of the context to switch to
        """
        if context_id not in self._contexts:
            raise ValueError(f"Context {context_id} does not exist")
        # get original active context
        original_context_id = self._active_context_id
        # log switch if it's a new context
        if original_context_id != context_id:
            logger.info(f"Switched from {original_context_id} to {context_id}") 

        # switch to new context
        self._active_context_id = context_id       
   


"""Input validation of user inputs for preprocessing"""

from typing import Literal, List
from findviz.logger_config import setup_logger

from findviz.viz.preprocess.fmri import PreprocessFMRIInputs
from findviz.viz.preprocess.timecourse import PreprocessTimecourseInputs
from findviz.viz.exception import PreprocessInputError
from findviz.viz.preprocess.validate import (
    validate_cutoff_exists, validate_cutoff_nyquist,
    validate_cutoff_order, validate_normalization,
    validate_tr_exists, validate_tr_positive,
    validate_fwhm_positive,
)

# Set up a logger for the app
logger = setup_logger(__name__)

class FMRIPreprocessInputValidator:
    """Class to validate user inputs for preprocessing"""
    def __init__(
        self,
        fmri_file_type: Literal['nifti', 'gifti']
    ):
        self.fmri_file_type = fmri_file_type

    def validate_preprocess_input(
        self, 
        input: PreprocessFMRIInputs
    ) -> bool:
        """Validate user inputs for fmri preprocessing

        Parameters
        ----------
        input : PreprocessFMRIInputs
            User inputs for preprocessing

        Returns
        -------
        bool
            True if inputs are valid, False otherwise
        """
        try:
            self._validate_any_input(input)
            # validate normalization parameters
            if input['normalize']:
                validate_normalization(input['mean_center'], input['zscore'])
            # validate filtering parameters
            if input['filter']:
                validate_tr_exists(input['tr'])
                validate_tr_positive(input['tr'])
                validate_cutoff_exists(input['low_cut'], input['high_cut'])
                # validate below nyquist frequency
                validate_cutoff_nyquist(
                    input['low_cut'], input['tr'], 'low cutoff'
                )
                validate_cutoff_nyquist(
                    input['high_cut'], input['tr'], 'high cutoff'
                )
                # validate cutoff order
                validate_cutoff_order(input['low_cut'], input['high_cut'])
            # validate smoothing parameters
            if input['smooth']:
                validate_fwhm_positive(input['fwhm'])

            return True
        # raise error to handle higher up
        except Exception as e:
            raise e
    
    def _validate_any_input(
        self,
        input: PreprocessFMRIInputs | PreprocessTimecourseInputs
    ) -> None:
        """Validate that at least one preprocessing parameter is passed"""
        input_missing = [
            input['detrend'] is False,
            input['mean_center'] is False,
            input['zscore'] is False,
            input['tr'] is None,
            input['low_cut'] is None,
            input['high_cut'] is None,
            input['fwhm'] is None
        ]
        if all(input_missing):
            logger.error('No preprocessing options selected')
            raise PreprocessInputError('No preprocessing options selected')


class TimecoursePreprocessInputValidator:
    """Class to validate user inputs for timecourse preprocessing"""
    def __init__(
        self,
    ):
        pass

        
    def validate_preprocess_input(
        self, 
        input: PreprocessTimecourseInputs,
        ts_labels: List[str]
    ) -> bool:
        """Validate user inputs for timecourse preprocessing

        Parameters
        ----------
        input : PreprocessTimecourseInputs
            User inputs for preprocessing
        ts_labels : List[str]
            Time course labels

        Returns
        -------
        bool
            True if inputs are valid, False otherwise
        """
        try:
            # validate that at least one time course is selected
            self._validate_ts_selected(ts_labels)
            # validate that at least one preprocessing parameter is passed
            self._validate_any_input(input)
            # validate normalization parameters
            if input['normalize']:
                validate_normalization(input['mean_center'], input['zscore'])
            # validate filtering parameters
            if input['filter']:
                validate_tr_exists(input['tr'])
                validate_tr_positive(input['tr'])
                validate_cutoff_exists(input['low_cut'], input['high_cut'])
                # validate below nyquist frequency
                validate_cutoff_nyquist(
                    input['low_cut'], input['tr'], 'low cutoff'
                )
                validate_cutoff_nyquist(
                    input['high_cut'], input['tr'], 'high cutoff'
                )
                # validate cutoff order
                validate_cutoff_order(input['low_cut'], input['high_cut'])

            return True
        # raise error to handle higher up
        except Exception as e:
            raise e
    
    def _validate_any_input(
        self,
        input: PreprocessTimecourseInputs
    ) -> None:
        """Validate that at least one preprocessing parameter is passed"""
        param_missing = [
            input['detrend'] is False,
            input['mean_center'] is False,
            input['zscore'] is False,
            input['tr'] is None,
            input['low_cut'] is None,
            input['high_cut'] is None,
        ]
        
        # check if any parameters are missing
        if all(param_missing):
            raise PreprocessInputError('No preprocessing options selected')
    
    def _validate_ts_selected(
        self,
        ts_labels: List[str]
    ) -> None:
        """Validate that at least one time course is selected"""
        if ts_labels is None or len(ts_labels) < 1:
            raise PreprocessInputError('No time courses selected')



"""
FIND Viewer: FMRI Interactive Navigation and Discovery Viewer

This package provides tools for visualizing and discovering patterns in fMRI data.
"""

from flask import Flask
from findviz.logger_config import setup_logger
from findviz.viz.io.cache import Cache

# Import core findviz module
from findviz import viz

# Set up package-level logger
logger = setup_logger(__name__)


def create_app(clear_cache=False, testing=False):
    """Application factory function."""
    app = Flask(__name__)

    # Clean up any existing cache on startup
    if clear_cache:
        cache = Cache()
        if cache.exists():
            cache.clear()
            logger.info("Cache cleared on startup")
    
    if testing:
        app.config.update({
            "TESTING": True,
            "WTF_CSRF_ENABLED": False,
        })
        
    # Import the blueprints
    from findviz.routes.file import file_bp
    from findviz.routes.main import main_bp
    from findviz.routes.viewer.analysis import analysis_bp
    from findviz.routes.viewer.color import color_bp
    from findviz.routes.viewer.data import data_bp
    from findviz.routes.viewer.io import io_bp
    from findviz.routes.viewer.plot import plot_bp
    from findviz.routes.viewer.preprocess import preprocess_bp
    from findviz.routes.viewer.logs import logs_bp


    # Register the blueprints
    app.register_blueprint(file_bp)
    app.register_blueprint(main_bp)
    app.register_blueprint(analysis_bp)
    app.register_blueprint(color_bp)
    app.register_blueprint(data_bp)
    app.register_blueprint(io_bp)
    app.register_blueprint(plot_bp)
    app.register_blueprint(preprocess_bp)
    app.register_blueprint(logs_bp)

    return app

# Version information
__version__ = '0.1.0'

__all__ = [
    # Core functionality
    'viz',

    # Utilities
    'setup_logger',
    
    # Version
    '__version__'
]



"""
File I/O utilities
"""
import csv
import os

from typing import List, Literal, Iterator, Optional, Union
from io import TextIOWrapper

from werkzeug.datastructures import FileStorage


def get_filename(file_input: Union[str, 'FileStorage']) -> str:
    """Get filename from either string path or FileStorage object.
    
    Parameters
    ----------
    file_input : Union[str, FileStorage]
        Either a string file path or FileStorage object
        
    Returns
    -------
    str
        The filename
    """
    if isinstance(file_input, str):
        return os.path.basename(file_input)
    else:  # FileStorage object
        return file_input.filename
    

def get_file_ext(fp: str) -> str:
    """
    get file ext from file path
    """
    base, ext = os.path.splitext(fp)
    return ext


def parse_nifti_file_ext(fp: str) -> Optional[Literal['.nii', '.nii.gz']]:
    """
    get file extension from nifti file path (.nii or .nii.gz). If the file
    is not a nifti file extension return None.
    """
    base, ext = os.path.splitext(fp)
    if ext == '.gz':
        base, ext = os.path.splitext(base)
        if ext == '.nii':
            return '.nii.gz'
        else:
            return
    elif ext == '.nii':
        return '.nii'
    else:
        return

def get_csv_reader(
    file,
    delimiter: str = ',',
    method=Literal['cli', 'browser']
) -> Iterator[List]:
    """
    Return csv reader with specified delimeter based on method of upload.
    Filestream for browser upload.
    """
    if method == 'cli':
        try:
            reader = csv.reader(file, delimiter = delimiter)
        # propagate error to parent function
        except Exception as e:
            raise e
    elif method == 'browser':
        try:
            file_stream = TextIOWrapper(file.stream, encoding='utf-8-sig')
            reader = csv.reader(file_stream, delimiter=delimiter)
        # propagate error to parent function
        except Exception as e:
            raise e
        
    return reader


"""
Preprocessing functions for functional MRI data
"""

from typing import Literal, Optional, TypedDict, Tuple

import nibabel as nib

from findviz.viz.transforms import (
    array_to_gifti, gifti_to_array, 
    nifti_to_array_masked, array_to_nifti_masked,
)
from findviz.viz.preprocess import utils
from findviz.viz.exception import NiftiMaskError


class PreprocessFMRIInputs(TypedDict):
    """
    Inputs for preprocess_fmri
    """
    normalize: bool
    filter: bool
    smooth: bool
    detrend: bool
    mean_center: bool
    zscore: bool
    tr: float
    low_cut: float
    high_cut: float
    fwhm: float


def preprocess_fmri(
    file_type: Literal['nifti', 'gifti'],
    inputs: PreprocessFMRIInputs,
    func_img: Optional[nib.Nifti1Image] = None,
    mask_img: Optional[nib.Nifti1Image] = None,
    left_func_img: Optional[nib.GiftiImage] = None,
    right_func_img: Optional[nib.GiftiImage] = None,
) -> nib.Nifti1Image | Tuple[nib.GiftiImage, nib.GiftiImage]:
    """
    Preprocess functional MRI data. Validation of preprocessing parameters
    is performed in the validate module.

    Arguments:
    ----------
        file_type: Literal['nifti', 'gifti']
            Type of file to preprocess
        inputs: PreprocessFMRIInputs
            Inputs for preprocess_fmri
        func_img: nib.Nifti1Image
            NIfTI image to preprocess. Only required if file_type is 'nifti'
        mask_img: nib.Nifti1Image
            Mask image to use for preprocessing. 
            Only required if file_type is 'nifti'
        left_func_img: nib.GiftiImage
            Gifti image containing left hemisphere functional data. 
            Only required if file_type is 'gifti'
        right_func_img: nib.GiftiImage
            Gifti image containing right hemisphere functional data. 
            Only required if file_type is 'gifti'
        
    Returns:
    --------
        nib.Nifti1Image | Tuple[nib.GiftiImage, nib.GiftiImage]
            Preprocessed functional MRI data. Either a single nib.Nifti1Image, 
            or a tuple of nib.GiftiImage objects (left and right hemispheres).
    """
    if file_type == 'nifti' and func_img is None:
        raise ValueError("func_img is required if file_type is 'nifti'")
    elif (file_type == 'gifti') and (left_func_img is None) and (right_func_img is None):
        raise ValueError(
            "left_func_img or right_func_img are required if file_type is 'gifti'"
        )

    # check if mask is provided for nifti processing
    if file_type == 'nifti' and mask_img is None:
        raise NiftiMaskError(
            message="A brain mask is required for nifti preprocessing",
        )

    # convert functional image to array
    if file_type == 'nifti':
        func_array = nifti_to_array_masked(func_img, mask_img)
    elif file_type == 'gifti':
        if left_func_img is not None and right_func_img is not None:
            both_hemispheres = True
        else:
            both_hemispheres = False
        func_array, split_index = gifti_to_array(
            left_func_img, right_func_img
        )

    # linear detrend
    if inputs['detrend']:
        func_array = utils.linear_detrend(func_array)

     # butterworth filter
    if inputs['filter']:
        # convert tr to hz
        tr_hz = utils.tr_to_hz(inputs['tr'])
        func_array = utils.butterworth_filter(
            func_array, 
            tr_hz, 
            inputs['low_cut'], 
            inputs['high_cut'], 
        )

    # mean center
    if inputs['mean_center']:
        func_array = utils.mean_center(func_array)

    # zscore
    if inputs['zscore']:
        func_array = utils.z_score(func_array)

    # 3d gaussian smooth (only for nifti)
    if inputs['smooth']:
        if file_type == 'nifti':
            # need to convert back to nifti before smoothing 
            func_img_prep = array_to_nifti_masked(func_array, mask_img)
            func_img_prep = utils.nifti_smooth(func_img_prep, inputs['fwhm'])
        elif file_type == 'gifti':
            raise NotImplementedError("Smoothing is not supported for Gifti files")

    # convert array back to nifti or gifti
    if file_type == 'nifti':
        if inputs['smooth']:
            return func_img_prep    
        else:
            return array_to_nifti_masked(func_array, mask_img)
    elif file_type == 'gifti':
        left_func_img_prep = None
        right_func_img_prep = None
        # if both hemispheres are provided, split the array
        if both_hemispheres:
            left_gii, right_gii = array_to_gifti(
                func_array, both_hemispheres, split_index
                )
            left_func_img_prep = left_gii
            right_func_img_prep = right_gii
        # if only one hemisphere is provided, assign the array to left or right
        else:
            gii_hemisphere = array_to_gifti(func_array, both_hemispheres=False)
            if left_func_img is not None:
                left_func_img_prep = gii_hemisphere
            elif right_func_img is not None:
                right_func_img_prep = gii_hemisphere

        return left_func_img_prep, right_func_img_prep





# Classes for scaling time courses via a constant or scale shift
from typing import List, Union

import numpy as np

class SignalScaler:
    """Class for scaling time courses via a scale shift. Tracks 
    history of transformations and allows for reversing the transformation."""
    def __init__(self):
        self.scale_history = []  # Store scaling factors

    def clear_history(self) -> None:
        """Clears the scale history."""
        self.scale_history = []
    
    def set_history(self, history: List[float]) -> None:
        """Sets the scale history directly"""
        self.scale_history = history

    def scale(
        self, 
        signals: List[List[float]], 
        scale_unit: float
    ) -> List[List[float]]:
        """Scales multiple signals' variance while preserving their means and stores the transformation.
        
        Arguments:
            signals: List of signals to scale
            scale_unit: Amount to scale the signals by
            
        Returns:
            List of scaled signals in same order as input
        """
        # Convert scale_unit to multiplicative factor
        # For increase: multiply by (1 + scale_unit)
        # For decrease: multiply by 1/(1 + scale_unit)
        scale_factor = 1 + scale_unit if scale_unit >= 0 else 1/(1 + abs(scale_unit))
        
        scaled_signals = []
        for signal in signals:
            mean = np.mean(signal)
            scaled_signal = [scale_factor * (s - mean) + mean for s in signal]
            scaled_signals.append(scaled_signal)
            
        self.scale_history.append(scale_factor)
        return scaled_signals

    def reverse(
        self, 
        signals: List[List[float]]
    ) -> List[List[float]]:
        """Reverses the last scaling transformation."""
        if not self.scale_history:
            print("No transformations to reverse.")
            return signals
        
        last_scale = self.scale_history.pop()  # Get the last applied scale
        reversed_signals = []
        for signal in signals:
            mean = np.mean(signal)
            reversed_signal = [(s - mean) / last_scale + mean for s in signal]
            reversed_signals.append(reversed_signal)

        return reversed_signals

    def reset(
        self, 
        signals: List[List[float]]
    ) -> List[List[float]]:
        """Resets the signal back to the original state."""
        # If no scale history, return original signals
        if not self.scale_history:
            return signals
        
        # reverse all scale transformations
        for _ in range(len(self.scale_history)):
            signals = self.reverse(signals)
        
        # check scale history is empty
        if self.scale_history:
            raise ValueError("Scale history is not empty after resetting.")
        
        return signals


class SignalShifter:
    """
    Class for shifting time courses via a constant shift. Tracks 
    history of transformations and allows for reversing the transformation.
    """
    def __init__(self):
        self.shift_history = []
    
    def clear_history(self) -> None:
        """Clears the shift history."""
        self.shift_history = []
    
    def set_history(self, history: List[float]) -> None:
        """Sets the shift history directly"""
        self.shift_history = history

    def shift(
        self, 
        signals: List[List[float]], 
        shift_amount: float
    ) -> List[List[float]]:
        """Shifts the signal by a constant amount and stores the transformation."""
        shifted_signals = []
        for signal in signals:
            shifted_signal = [s + shift_amount for s in signal]
            shifted_signals.append(shifted_signal)
        self.shift_history.append(shift_amount)
        return shifted_signals

    def reverse(
        self, 
        signals: List[List[float]]
    ) -> List[List[float]]:
        """Reverses the last shift transformation."""
        if not self.shift_history:
            print("No transformations to reverse.")
            return signals
        
        last_shift = self.shift_history.pop()  # Get the last applied shift
        reversed_signals = []
        for signal in signals:
            reversed_signal = [s - last_shift for s in signal]
            reversed_signals.append(reversed_signal)
        return reversed_signals

    def reset(
        self, 
        signals: List[List[float]]
    ) -> List[List[float]]:
        """Resets the signal back to the original state."""
        # reverse all shift transformations
        for _ in range(len(self.shift_history)):
            signals = self.reverse(signals)
        
        # check shift history is empty
        if self.shift_history:
            raise ValueError("Shift history is not empty after resetting.")
        
        return signals



import logging
import os
from pathlib import Path
import time

def setup_logger(name=__name__, disable_file_logging=False):
    """
    Logger set up for findviz app
    
    Parameters
    ----------
    name : str
        Logger name
    disable_file_logging : bool
        If True, only console logging will be enabled (useful for testing)
    """
    # Create a logger
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    
    # Prevent duplicate handlers
    if not logger.handlers:
        # Create a formatter
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        )
        
        # Create a console handler
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        console_handler.setLevel(logging.DEBUG)
        logger.addHandler(console_handler)

        # Create a file handler if not disabled
        if not disable_file_logging:
            # Get the current working directory and ensure logs directory exists
            current_dir = os.getcwd()
            log_dir = os.path.join(current_dir, 'logs')
            os.makedirs(log_dir, exist_ok=True)
            
            # Create a run-specific log file with timestamp
            run_timestamp = time.strftime("%Y%m%d-%H%M%S")
            log_file_path = os.path.join(log_dir, f'app-run-{run_timestamp}.log')
            
            # Set up a file handler for this run
            file_handler = logging.FileHandler(
                filename=log_file_path,
                encoding='utf-8'
            )
            
            file_handler.setFormatter(formatter)
            file_handler.setLevel(logging.INFO)
            logger.addHandler(file_handler)
    
    return logger


"""
FMRI Data Transforms

Includes conversions between nifti/gifti image formats and numpy arrays and vice versa.

"""

from typing import Optional, Tuple

import nibabel as nib
import numpy as np

from nilearn import masking
from nibabel.gifti import GiftiDataArray

def array_to_nifti(
    array: np.ndarray, 
    affine: np.ndarray, 
    header: nib.Nifti1Header
) -> nib.Nifti1Image:
    """
    Convert numpy array to nibabel image

    Arguments:
    ----------
        array: numpy array
        affine: affine matrix
        header: nifti header
    
    Returns:
    --------
        nifti_img: nibabel image
    """
    return nib.Nifti1Image(array, affine, header)


def array_to_nifti_masked(array: np.ndarray, mask_img: nib.Nifti1Image) -> nib.Nifti1Image:
    """
    Convert 2d numpy array to nifti image

    Arguments:
    ----------
        array: 2d numpy array
        mask_img: mask nifti image (0 = exclude, 1 = include)
    
    Returns:
    --------
        nifti_img: nibabel image
    """
    nifti_img = masking.unmask(array, mask_img)
    return nifti_img


def nifti_to_array(nifti_img: nib.Nifti1Image) -> np.ndarray:
    """
    Convert nifti image to 3d or 4d numpy array

    Arguments:
    ----------
        nifti_img: nibabel image
    
    Returns:
    --------
        array: numpy array
    """ 
    return nifti_img.get_fdata()


def nifti_to_array_masked(nifti_img: nib.Nifti1Image, mask_img: nib.Nifti1Image) -> np.ndarray:
    """
    Convert nifti image to 2d numpy array. Voxels are arranged along columns 
    and time points along rows.

    Arguments:
    ----------
        nifti_img: nifti image
        mask_img: mask nifti image (0 = exclude, 1 = include)
    
    Returns:
    --------
        array: 2d numpy array
    """
    nifti_2d = masking.apply_mask(nifti_img, mask_img)
    return nifti_2d


def gifti_to_array(
    left_gifti: Optional[nib.GiftiImage] = None, 
    right_gifti: Optional[nib.GiftiImage] = None
) -> Tuple[np.ndarray, int]:
    """
    Convert gifti image to 2d numpy array. If both hemispheres are provided, 
    they are concatenated along the columns with left hemisphere first.
    The split index is returned as the second argument.
    """
    split_index = None
    if left_gifti is None and right_gifti is None:
        raise ValueError("No gifti images provided")
    
    if left_gifti is None:
        return _gifti_extract_data(right_gifti), split_index
    if right_gifti is None:
        return _gifti_extract_data(left_gifti), split_index
    
    # Get the length of left hemisphere data
    split_index = len(left_gifti.darrays[0].data)
    # Concatenate the data from both hemispheres
    concat_data = np.hstack([
        _gifti_extract_data(left_gifti),
        _gifti_extract_data(right_gifti)
    ])
    return concat_data, split_index


def array_to_gifti(
    array: np.ndarray,
    both_hemispheres: bool = False,
    split_index: Optional[int] = None
) -> Tuple[nib.GiftiImage, nib.GiftiImage] | nib.GiftiImage:
    """
    Convert 2d numpy array to gifti image. If both hemispheres are provided, 
    the array is split along the columns with left hemisphere first. If 
    both_hemispheres is False, the array is added to a single gifti image. If
    both_hemispheres is True, the split index is required.

    Arguments:
    ----------
        array: 2d numpy array (concatenated if both_hemispheres is True)
        both_hemispheres: whether to return both hemispheres
        split_index: split index. Required if both_hemispheres is True.
    
    Returns:
    --------
        gifti_img: nibabel gifti image (left or right) or both hemispheres - with
        left hemisphere first.

    """
    if both_hemispheres and split_index is None:
        raise ValueError("Split index is required for concatenated array")
    
    # create gifti images
    gifti_img = nib.GiftiImage()
    left_gifti = nib.GiftiImage()
    right_gifti = nib.GiftiImage()

    # loop over rows of array (row = time point)
    for row_i in range(array.shape[0]):
        if both_hemispheres:
            # split array into left and right hemispheres
            left_gifti.add_gifti_data_array(
                GiftiDataArray(array[row_i, :split_index], datatype=16)
            )
            right_gifti.add_gifti_data_array(
                GiftiDataArray(array[row_i, split_index:], datatype=16)
            )
        else:
            # add all rows to gifti image
            gifti_img.add_gifti_data_array(
                GiftiDataArray(array[row_i,:], datatype=16)
            )

    if both_hemispheres:
        return left_gifti, right_gifti
    else:
        return gifti_img


def _gifti_extract_data(gifti_img: nib.GiftiImage) -> np.ndarray:
    """
    Extract data from gifti image
    """
    gifti_data = np.vstack([darray.data for darray in gifti_img.darrays])
    return gifti_data





"""Validate user inputs for preprocessing"""

from typing import Literal, Tuple

from findviz.viz.exception import PreprocessInputError


def validate_cutoff_exists(low_cut: float, high_cut: float) -> bool:
    """Validate at least one cutoff frequency exists
    
    Parameters
    ----------
    low_cut : float
        Low cutoff frequency in Hz
    high_cut : float
        High cutoff frequency in Hz
        
    Returns
    -------
    bool
        Validation status
    """
    if (low_cut is None or low_cut == '') and (high_cut is None or high_cut == ''):
        error_msg = 'at least lowCut or highCut (in Hz) must be provided'
        raise PreprocessInputError(error_msg, 'filtering')
    return True

def validate_cutoff_nyquist(
    freq: float, 
    tr: float, 
    cutoff_type: Literal['low', 'high']
) -> bool:
    """Validate cutoff frequency is below Nyquist frequency
    
    Parameters
    ----------
    freq : float
        Cutoff frequency in Hz
    tr : float
        Repetition time in seconds
    cutoff_type : str
        Type of cutoff ('lowCut' or 'highCut')
        
    Returns
    -------
    bool
        Validation status
    """
    if freq is not None and freq != '':
        nyquist_freq = (1/tr)/2
        if float(freq) > nyquist_freq:
            error_msg = f'{cutoff_type} must be less than Nyquist frequency: {nyquist_freq}'
            raise PreprocessInputError(error_msg, 'filtering')
    return True

def validate_cutoff_order(low_cut: float, high_cut: float) -> bool:
    """Validate low cutoff is less than high cutoff
    
    Parameters
    ----------
    low_cut : float
        Low cutoff frequency in Hz
    high_cut : float
        High cutoff frequency in Hz
        
    Returns
    -------
    bool
        Validation status
    """
    if (low_cut is not None and low_cut != '') and (high_cut is not None and high_cut != ''):
        if float(low_cut) >= float(high_cut):
            error_msg = 'lowCut must be less than highCut'
            raise PreprocessInputError(error_msg, 'filtering')
    return True

def validate_normalization(mean_center: bool, zscore: bool) -> bool:
    """Validate normalization options
    
    Parameters
    ----------
    mean_center : bool
        Whether to mean center the data
    zscore : bool
        Whether to z-score the data
        
    Returns
    -------
    bool
        Validation status
    """
    if not mean_center and not zscore:
        error_msg = 'If normalization is enabled, mean-center or z-score option must be selected'
        raise PreprocessInputError(error_msg, 'normalization')
    return True

def validate_fwhm_positive(fwhm: float) -> bool:
    """Validate FWHM is positive
    
    Parameters
    ----------
    full_width_half_max : float
        FWHM in mm
        
    Returns
    -------
    bool
        Validation status
    """
    if fwhm < 0:
        error_msg = 'FWHM values must be positive'
        raise PreprocessInputError(error_msg, 'smoothing')
    return True

def validate_tr_exists(tr: float) -> bool:
    """Validate TR exists and is not empty
    
    Parameters
    ----------
    tr : float
        Repetition time in seconds
        
    Returns
    -------
    bool
        Validation status
    """
    if tr is None or tr == '':
        error_msg = 'TR (repetition time) of functional time courses must be provided'
        raise PreprocessInputError(error_msg, 'filtering')
    return True

def validate_tr_positive(tr: float) -> bool:
    """Validate TR is positive
    
    Parameters
    ----------
    tr : float
        Repetition time in seconds
        
    Returns
    -------
    bool
        Validation status
    """
    if tr < 0:
        error_msg = 'TR (repetition time) must be a positive number'
        raise PreprocessInputError(error_msg, 'filtering')
    return True



"""
Shared resources for the application.
"""
from findviz.logger_config import setup_logger

logger = setup_logger(__name__)

from findviz.viz.viewer.data_manager import DataManager

# Create the single shared instance
try:
    data_manager = DataManager()
except Exception as e:
    logger.error("Error initializing data manager: %s", str(e), exc_info=True)
    raise





"""
Distance analysis of fmri time points
"""
from typing import List, Tuple

import numpy as np

from scipy.spatial.distance import cdist


class Distance:
    """Distance analysis of fmri time points"""

    def __init__(
        self,
        distance_metric: str,
    ):
        self.distance_metric = distance_metric
    
    def calculate_distance(self, time_point: int, fmri_data: np.ndarray) -> np.ndarray:
        """Calculate distance between fmri time point and rest of time points"""
        dist = cdist(
            fmri_data[time_point,:][np.newaxis,:], 
            fmri_data, 
            metric=self.distance_metric
        )
        return np.squeeze(dist)



"""
Viewer IO routes
"""

from io import BytesIO

from flask import Blueprint, request, send_file, make_response

from findviz.logger_config import setup_logger
from findviz.routes.utils import (
    Routes
)
from findviz.routes.shared import data_manager

# Set up a logger for the app
logger = setup_logger(__name__)
io_bp = Blueprint('io', __name__)


@io_bp.route(Routes.SAVE_SCENE.value, methods=['POST'])
def save_scene() -> dict:
    """
    Save the current scene
    """

    # Get serialized data from DataManager
    serialized_data = data_manager.save()
    
    # Create a BytesIO object from the serialized data
    mem_file = BytesIO(serialized_data)
    mem_file.seek(0)
    
    # Create a response with the file data
    response = send_file(
        mem_file,
        mimetype='application/octet-stream',
        as_attachment=True,
        download_name='scene'
    )
    
    # Set headers to prevent caching
    response.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
    response.headers["Pragma"] = "no-cache"
    response.headers["Expires"] = "0"

    # return response
    return response



"""
Utilities for handling nifti file uploads
"""
import gzip

from enum import Enum
from io import BytesIO
from pathlib import Path
from typing import Literal, Optional, Dict, Union

import nibabel as nib

from flask import request
from nilearn.image import reorder_img
from werkzeug.datastructures import FileStorage

from findviz.viz import exception
from findviz.viz.io import validate
from findviz.viz.io import utils

# Type aliases
FilePath = Union[str, Path]
FileUploadDict = Dict[str, Optional[Union[FilePath, 'FileStorage']]]
NiftiDict = Dict[str, Optional[nib.Nifti1Image]]

# Expected file upload inputs
class NiftiFiles(Enum):
    FUNC = 'nii_func'
    ANAT = 'nii_anat'
    MASK = 'nii_mask'

# Browser fields type definition
browser_fields: Dict[str, str] = {
    NiftiFiles.FUNC.value: 'nifti-func',
    NiftiFiles.ANAT.value: 'nifti-anat',
    NiftiFiles.MASK.value: 'nifti-mask',
}

class NiftiUpload:
    """
    Class for handling nifti file uploads.

    Attributes
    ----------
    method : Literal['cli', 'browser']
        The method used for file uploads
    """

    def __init__(
        self,
        method: Literal['cli', 'browser']
    ):
        self.method = method

    def upload(
        self,
        fmri_files: Optional[Dict[str, Union[str, Path]]] = None,
    ) -> NiftiDict:
        """
        Get and validate nifti files uploaded from cli or browser.

        Parameters
        ----------
        fmri_files : Optional[Dict[str, FilePath]]
            Dictionary of Nifti file paths: 
            {'nii_func': path, 'nii_anat': path, 'nii_mask': path}

        Returns
        -------
        NiftiDict
            Dictionary containing validated Nifti images

        Raises
        ------
        FileInputError
            If required files are missing or have invalid extensions
        FileUploadError
            If there are issues reading the files
        FileValidationError
            If files fail validation checks
        """
        if self.method == 'cli':
            file_uploads = fmri_files
        elif self.method == 'browser':
            file_uploads = self._get_browser_input()

        # nifti file is not optional
        if file_uploads[NiftiFiles.FUNC.value] is None:
            raise exception.FileInputError(
                'Please upload a Nifti file (.nii or .nii.gz). for '
                'functional file',
                exception.ExceptionFileTypes.NIFTI.value, self.method,
                [browser_fields[NiftiFiles.FUNC.value]]
            )
        
        # validate func file extension
        if not validate.validate_nii_ext(
            utils.get_filename(file_uploads[NiftiFiles.FUNC.value])
        ):
            raise exception.FileInputError(
                'File extension is not recognized. Please upload a Nifti file '
                 ' (.nii or .nii.gz). for functional file',
                exception.ExceptionFileTypes.NIFTI.value, self.method,
                [browser_fields[NiftiFiles.FUNC.value]]
            )

        # Initialize output dictionary
        nifti_out = {
            NiftiFiles.FUNC.value: None,
            NiftiFiles.ANAT.value: None,
            NiftiFiles.MASK.value: None
        }
        # Load and validate functional
        try:
            nii_func = read_nii(file_uploads[NiftiFiles.FUNC.value], self.method)
        except Exception as e:
            raise exception.FileUploadError(
                'Error in reading functional nifti file. Please check format.',
                exception.ExceptionFileTypes.NIFTI.value, self.method,
                [browser_fields[NiftiFiles.FUNC.value]]
            ) from e
        # validate functional is 4d
        if not validate.validate_nii_4d(nii_func):
            raise exception.FileValidationError(
                "The functional file must have 4-dimensions (x,y,z,time). "
                 " Please check file.",
                validate.validate_nii_4d.__name__,
                exception.ExceptionFileTypes.NIFTI.value,
                [browser_fields[NiftiFiles.FUNC.value]]
            )

        # Reorient to RAS
        nii_func = reorder_img(nii_func)
        
        nifti_out[NiftiFiles.FUNC.value] = nii_func

        # Load and validate anatomical
        if file_uploads[NiftiFiles.ANAT.value] is not None:
            # validate anat file extension
            if not validate.validate_nii_ext(
                utils.get_filename(file_uploads[NiftiFiles.ANAT.value])
            ):
                raise exception.FileInputError(
                    'File extension is not recognized. Please upload a Nifti file '
                    ' (.nii or .nii.gz). for anatomical file',
                    exception.ExceptionFileTypes.NIFTI.value, self.method,
                    [browser_fields[NiftiFiles.ANAT.value]]
                )
            
            try:
                nii_anat = read_nii(file_uploads[NiftiFiles.ANAT.value], self.method)
            except Exception as e:
                raise exception.FileUploadError(
                    'Error in reading anatomical nifti file',
                    exception.ExceptionFileTypes.NIFTI.value, self.method,
                    [browser_fields[NiftiFiles.ANAT.value]]
                ) from e
            # validate anat is 3d
            if not validate.validate_nii_3d(nii_anat):
                raise exception.FileValidationError(
                    "The anat file must have only 3-dimensions (x,y,z). Please "
                    "check file.",
                    validate.validate_nii_3d.__name__,
                    exception.ExceptionFileTypes.NIFTI.value,
                    [browser_fields[NiftiFiles.ANAT.value]]
                )
            # validate that anat and func voxel dimension lengths are consistent
            if not validate.validate_nii_same_dim_len(nii_anat, nii_func):
                raise exception.FileValidationError(
                    "The anat file does not have the same voxel dimension "
                    "lengths as the functional file. Please check file.",
                    validate.validate_nii_same_dim_len.__name__,
                    exception.ExceptionFileTypes.NIFTI.value,
                    [browser_fields[NiftiFiles.ANAT.value]]
                )
            
            # Reorient to RAS
            nii_anat = reorder_img(nii_anat)
            nifti_out[NiftiFiles.ANAT.value] = nii_anat
        
        # load and validate mask 
        if file_uploads[NiftiFiles.MASK.value] is not None:
             # validate mask file extension
            if not validate.validate_nii_ext(
                utils.get_filename(file_uploads[NiftiFiles.MASK.value])
            ):
                raise exception.FileInputError(
                    'File extension is not recognized. Please upload a Nifti file '
                    ' (.nii or .nii.gz). for mask file',
                    exception.ExceptionFileTypes.NIFTI.value, self.method,
                    [browser_fields[NiftiFiles.MASK.value]]
                )
            
            try:
                nii_mask = read_nii(file_uploads[NiftiFiles.MASK.value], self.method)
            except Exception as e:
                raise exception.FileUploadError(
                    'Error in reading mask nifti file',
                    exception.ExceptionFileTypes.NIFTI.value, self.method,
                    [browser_fields[NiftiFiles.MASK.value]]
                ) from e
            # validate mask is 3d
            if not validate.validate_nii_3d(nii_mask):
                raise exception.FileValidationError(
                    "The mask file must have only 3-dimensions (x,y,z). Please "
                    "check file.",
                    validate.validate_nii_3d.__name__,
                    exception.ExceptionFileTypes.NIFTI.value,
                    [browser_fields[NiftiFiles.MASK.value]]
                )
            # validate that mask and func voxel dimension lengths are consistent
            if not validate.validate_nii_same_dim_len(nii_mask, nii_func):
                raise exception.FileValidationError(
                    "The mask file does not have the same voxel dimension "
                    "lengths as the functional file. Please check file.",
                    validate.validate_nii_same_dim_len.__name__,
                    exception.ExceptionFileTypes.NIFTI.value,
                    [browser_fields[NiftiFiles.MASK.value]]
                )
            # validate that mask only contains 1s and 0s
            if not validate.validate_nii_brain_mask(nii_mask):
                raise exception.FileValidationError(
                    "The mask file must only contain 1s and 0s. Please check file.",
                    validate.validate_nii_brain_mask.__name__,
                    exception.ExceptionFileTypes.NIFTI.value,
                    [browser_fields[NiftiFiles.MASK.value]]
                )
            # Reorient to RAS
            nii_mask = reorder_img(nii_mask)
            nifti_out[NiftiFiles.MASK.value] = nii_mask

        return nifti_out

    @staticmethod
    def _get_browser_input() -> dict:
        """
        Get file uploads from browser request.

        Returns
        -------
        FileUploadDict
            Dictionary containing file upload objects
        """
        nifti_file = request.files.get(NiftiFiles.FUNC.value)
        anatomical_file = request.files.get(NiftiFiles.ANAT.value)
        mask_file = request.files.get(NiftiFiles.MASK.value)

        file_upload = {
            NiftiFiles.FUNC.value: nifti_file,
            NiftiFiles.ANAT.value: anatomical_file,
            NiftiFiles.MASK.value: mask_file
        }
        return file_upload


def read_nii(
    file: Union[FilePath, 'FileStorage'],
    method: Literal['cli', 'browser']
) -> nib.Nifti1Image:
    """
    Read nifti file based on method of upload.

    Parameters
    ----------
    file : Union[FilePath, FileStorage]
        Either a nifti file from the browser or full path to file from CLI
    method : Literal['cli', 'browser']
        Whether the file was uploaded through browser or CLI

    Returns
    -------
    nib.Nifti1Image
        Loaded nibabel image

    Raises
    ------
    Exception
        If there are issues reading the file
    """
    if method =='cli':
        try:
            nifti_img = nib.load(file)
        # raise generic exception to be handled higher in stack
        except Exception as e:
            raise e
    elif method == 'browser':
        # get file extension
        ext = utils.parse_nifti_file_ext(file.filename)
        try:
            nifti_img = _read_nii_browser(file, ext=ext)
        except Exception as e:
            raise e
    return nifti_img


def _read_nii_browser(
    file: 'FileStorage',
    ext: Literal['.nii', '.nii.gz']
) -> nib.Nifti1Image:
    """
    Read .nii or .nii.gz compressed nifti file from browser.

    Parameters
    ----------
    file : FileStorage
        File upload object from browser
    ext : Literal['.nii', '.nii.gz']
        File extension

    Returns
    -------
    nib.Nifti1Image
        Loaded nibabel image

    Raises
    ------
    Exception
        If there are issues reading the file
    """
    file_stream = BytesIO(file.read())
    try:
        if ext == '.nii.gz':
            with gzip.GzipFile(fileobj=file_stream, mode="rb") as d_stream:
                nifti_data = BytesIO(d_stream.read())
            file_stream.close()
            nifti_img = nib.Nifti1Image.from_bytes(nifti_data.read())
        elif ext == '.nii':
            nifti_img = nib.Nifti1Image.from_bytes(file_stream.read())
    # raise generic exception to be handled higher in stack
    except Exception as e:
        raise e
    return nifti_img









